import re
from abc import ABC, abstractmethod
from typing import List, Tuple, Union

from ..utils.custom_types import NERSample, Sample, SequenceClassificationSample


class BaseFormatter(ABC):
    """Abstract base class for defining formatter classes.

    Subclasses should implement the static methods `to_csv` and `to_conll`.
    """

    @staticmethod
    @abstractmethod
    def to_csv(sample: Sample):
        """Converts a custom type to a CSV string.

        Args:
            sample (Sample): The custom type to convert.

        Returns:
            The CSV string representation of the custom type.

        Raises:
            NotImplementedError: This method should be implemented by the subclass.
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def to_conll(sample: Sample):
        """Converts a custom type to a CoNLL string.

        Args:
            sample (Sample): The custom type to convert.

        Returns:
            The CoNLL string representation of the custom type.

        Raises:
            NotImplementedError: This method should be implemented by the subclass.
        """
        raise NotImplementedError()


class Formatter:
    """Formatter class for converting between custom types and different output formats.

    This class uses the `to_csv` and `to_conll` methods of subclasses of `BaseFormatter`
    to perform the conversions. The appropriate subclass is selected based on the
    type of the expected results in the `sample` argument.
    """

    @staticmethod
    def process(sample: Sample, output_format: str, *args, **kwargs):
        """Method to format the sample into the desired format

        Args:
            sample (Sample):
                The input sample to convert.
            output_format (str):
                The output format to convert to, either "csv" or "conll".
            *args:
                Optional positional arguments to pass to the `to_csv` or `to_conll` methods.
            **kwargs:
                Optional keyword arguments to pass to the `to_csv` or `to_conll` methods.

        Returns:
            The output string in the specified format.

        Raises:
            NameError: If no formatter subclass is defined for the type of the expected results in the sample.

        """
        formats = {cls.__name__: cls for cls in BaseFormatter.__subclasses__()}
        class_name = type(sample.expected_results).__name__
        try:
            return getattr(formats[f"{class_name}Formatter"], f"to_{output_format}")(
                sample, *args, **kwargs
            )
        except KeyError:
            raise NameError(f"Class '{class_name}Formatter' not yet implemented.")


class SequenceClassificationOutputFormatter(BaseFormatter, ABC):
    """Formatter class for converting `SequenceClassificationOutput` objects to CSV.

    The `to_csv` method returns a CSV string representing the `SequenceClassificationOutput`
    object in the sample argument.
    """

    @staticmethod
    def to_csv(sample: SequenceClassificationSample) -> Tuple[str, str]:
        """Convert a SequenceClassificationSample object into a row for exporting.

        Args:
            sample (SequenceClassificationSample) : Sample object to convert.

        Returns:
            Tuple[str, str]:
                Row formatted as a list of strings.
        """
        if sample.test_case:
            return [sample.test_case, sample.expected_results.predictions[0].label]
        return [sample.original, sample.expected_results.predictions[0].label]


class NEROutputFormatter(BaseFormatter):
    """Formatter class for converting `NEROutput` objects to CSV and CoNLL.

    The `to_csv` method returns a CSV string representing the `NEROutput` object in the sample
    argument. The `to_conll` method returns a CoNLL string representing the `NEROutput` object.
    """

    @staticmethod
    def to_csv(
        sample: NERSample, delimiter: str = ",", temp_id: int = None
    ) -> Tuple[List[str], List[str], List[str], List[str]]:
        """Converts a NERSample to a CSV string.

        Args:
            sample (NERSample):
                The input sample containing the `NEROutput` object to convert.
            delimiter (str):
                The delimiter character to use in the CSV string.
            temp_id (int):
                A temporary ID to use for grouping entities by document.

        Returns:
            Tuple[List[str], List[str], List[str], List[str]]:
                The CSV or CoNLL string representation of the `NEROutput` object along with the document id
        """
        test_case = sample.test_case
        original = sample.original

        words = re.finditer(r"([^\s]+)", original)
        tokens, labels = [], []

        for word in words:
            tokens.append(word.group())
            match = sample.expected_results[word.group()]
            labels.append(match.entity if match is not None else "O")

        if test_case and sample.actual_results:
            test_case_words = re.finditer(r"([^\s]+)", test_case)
            test_case_tokens, test_case_labels = [], []

            for word in test_case_words:
                test_case_tokens.append(word.group())
                match = sample.actual_results[word.group()]
                test_case_labels.append(match.entity if match is not None else "O")

            assert len([token for token in test_case_tokens if token != "O"]) == len(
                sample.actual_results
            )
            return tokens, labels, test_case_tokens, test_case_labels
        return tokens, labels, [], []

    @staticmethod
    def to_conll(sample: NERSample, temp_id: int = None) -> Union[str, Tuple[str, str]]:
        """Converts a custom type to a CoNLL string.

        Args:
            sample (NERSample):
                The input sample containing the `NEROutput` object to convert.
            writing_mode (str):
                what to do with the expected results if present:
                - ignore: simply ignores the expected_results
                - append: the formatted expected_results to the original ones
                - separate: returns a formatted string for the original sentence and one for
                            the perturbed sentence

        Returns:
            The CoNLL string representation of the custom type.
        """

        text = ""
        test_case = sample.test_case
        original = sample.original
        if test_case:
            test_case_items = test_case.split()
            norm_test_case_items = test_case.lower().split()
            norm_original_items = original.lower().split()
            temp_len = 0
            for jdx, item in enumerate(norm_test_case_items):
                try:
                    if item in norm_original_items and jdx >= norm_original_items.index(
                        item
                    ):
                        oitem_index = norm_original_items.index(item)
                        j = sample.expected_results.predictions[oitem_index + temp_len]
                        if temp_id != j.doc_id and jdx == 0:
                            text += f"{j.doc_name}\n\n"
                            temp_id = j.doc_id
                        text += f"{test_case_items[jdx]} {j.pos_tag} {j.chunk_tag} {j.entity}\n"
                        norm_original_items.pop(oitem_index)
                        temp_len += 1
                    else:
                        o_item = sample.expected_results.predictions[jdx].span.word
                        letters_count = len(set(item) - set(o_item))
                        if (
                            len(norm_test_case_items) == len(original.lower().split())
                            or letters_count < 2
                        ):
                            tl = sample.expected_results.predictions[jdx]
                            text += f"{test_case_items[jdx]} {tl.pos_tag} {tl.chunk_tag} {tl.entity}\n"
                        else:
                            text += f"{test_case_items[jdx]} -X- -X- O\n"
                except IndexError:
                    text += f"{test_case_items[jdx]} -X- -X- O\n"

        else:
            for j in sample.expected_results.predictions:
                if temp_id != j.doc_id:
                    text += f"{j.doc_name}\n\n"
                    temp_id = j.doc_id
                text += f"{j.span.word} {j.pos_tag} {j.chunk_tag} {j.entity}\n"

        return text, temp_id



import asyncio
from abc import ABC, abstractmethod
from typing import List, Dict, Union

from langtest.modelhandler.modelhandler import ModelFactory
from langtest.utils.custom_types import (
    MaxScoreOutput,
    MaxScoreSample,
    MinScoreOutput,
    MinScoreSample,
    Sample,
)
from langtest.utils.util_metrics import calculate_f1_score


class BaseFairness(ABC):
    """Abstract base class for implementing accuracy measures.

    Attributes:
        alias_name (str): A name or list of names that identify the accuracy measure.

    Methods:
        transform(data: List[Sample], params: Dict) -> Union[List[MinScoreSample], List[MaxScoreSample]]:
            Transforms the input data into an output based on the implemented accuracy measure.
    """

    alias_name = None
    supported_tasks = ["ner", "text-classification"]

    @staticmethod
    @abstractmethod
    def transform(
        data: List[Sample], params: Dict
    ) -> Union[List[MinScoreSample], List[MaxScoreSample]]:
        """Abstract method that implements the computation of the given measure.

        Args:
            data (List[Sample]): The input data to be transformed.
            params (Dict): parameters for tests configuration
        Returns:
            Union[List[MinScoreSample], List[MaxScoreSample]]: The transformed data based on the implemented measure.
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    async def run(
        sample_list: List[MinScoreSample], categorised_data, **kwargs
    ) -> List[Sample]:
        """Computes the score for the given data.

        Args:
            sample_list (List[MinScoreSample]): The input data to be transformed.
            model (ModelFactory): The model to be used for the computation.

        Returns:
            List[MinScoreSample]: The transformed samples.
        """
        raise NotImplementedError()

    @classmethod
    async def async_run(cls, sample_list: List[Sample], model: ModelFactory, **kwargs):
        """Creates a task for the run method.

        Args:
            sample_list (List[Sample]): The input data to be transformed.
            model (ModelFactory): The model to be used for the computation.

        Returns:
            asyncio.Task: The task for the run method.

        """
        created_task = asyncio.create_task(cls.run(sample_list, model, **kwargs))
        return created_task


class MinGenderF1Score(BaseFairness):
    """Subclass of BaseFairness that implements the minimum F1 score.

    Attributes:
        alias_name (str): The name "min_f1" identifying the minimum F1 score.

    Methods:
        transform(data: List[Sample], params: Dict) -> List[MinScoreSample]:
            Transforms the input data into an output based on the minimum F1 score.
    """

    alias_name = ["min_gender_f1_score"]

    @classmethod
    def transform(
        cls, test: str, data: List[Sample], params: Dict
    ) -> List[MinScoreSample]:
        """Computes the minimum F1 score for the given data.

        Args:
            test (str): name of the test
            data (List[Sample]): The input data to be transformed.
            params (Dict): parameters for tests configuration.
        Returns:
            List[MinScoreSample]: The transformed data based on the minimum F1 score.
        """

        assert (
            test in cls.alias_name
        ), f"Parameter 'test' should be in: {cls.alias_name}, got '{test}'"

        if isinstance(params["min_score"], dict):
            min_scores = params["min_score"]
        elif isinstance(params["min_score"], float):
            min_scores = {
                "male": params["min_score"],
                "female": params["min_score"],
                "unknown": params["min_score"],
            }

        samples = []
        for key, val in min_scores.items():
            sample = MinScoreSample(
                original=None,
                category="fairness",
                test_type="min_gender_f1_score",
                test_case=key,
                expected_results=MinScoreOutput(min_score=val),
            )

            samples.append(sample)
        return samples

    @staticmethod
    async def run(
        sample_list: List[MinScoreSample], gendered_data, **kwargs
    ) -> List[MinScoreSample]:
        """Computes the minimum F1 score for the given data.

        Args:
            sample_list (List[MinScoreSample]): The input data to be transformed.
            model (ModelFactory): The model to be used for the computation.

        Returns:
            List[MinScoreSample]: The transformed samples.

        """
        progress = kwargs.get("progress_bar", False)
        for sample in sample_list:
            data = gendered_data[sample.test_case]
            if len(data[0]) > 0:
                macro_f1_score = calculate_f1_score(
                    data[0].to_list(), data[1].to_list(), average="macro", zero_division=0
                )
            else:
                macro_f1_score = 1

            sample.actual_results = MinScoreOutput(min_score=macro_f1_score)
            sample.state = "done"

            if progress:
                progress.update(1)
        return sample_list


class MaxGenderF1Score(BaseFairness):
    """Subclass of BaseFairness that implements the maximum F1 score.

    Attributes:
        alias_name (str): The name to be used in config.

    Methods:
        transform(data: List[Sample], params: Dict) -> List[MaxScoreSample]:
            Transforms the input data into an output based on the maximum F1 score.
    """

    alias_name = ["max_gender_f1_score"]

    @classmethod
    def transform(
        cls, test: str, data: List[Sample], params: Dict
    ) -> List[MaxScoreSample]:
        """Computes the maximum F1 score for the given data.

        Args:
            test (str): name of the test.
            data (List[Sample]): The input data to be transformed.
            params (Dict): parameters for tests configuration
        Returns:
            List[MaxScoreSample]: The transformed data based on the maximum F1 score.
        """
        assert (
            test in cls.alias_name
        ), f"Parameter 'test' should be in: {cls.alias_name}, got '{test}'"
        if isinstance(params["max_score"], dict):
            max_scores = params["max_score"]
        elif isinstance(params["max_score"], float):
            max_scores = {
                "male": params["max_score"],
                "female": params["max_score"],
                "unknown": params["max_score"],
            }

        samples = []
        for key, val in max_scores.items():
            sample = MaxScoreSample(
                original=None,
                category="fairness",
                test_type="max_gender_f1_score",
                test_case=key,
                expected_results=MaxScoreOutput(max_score=val),
            )

            samples.append(sample)
        return samples

    @staticmethod
    async def run(
        sample_list: List[MaxScoreSample], gendered_data, **kwargs
    ) -> List[MaxScoreSample]:
        """Computes the maximum F1 score for the given data.

        Args:
            sample_list (List[MaxScoreSample]): The input data to be transformed.
            model (ModelFactory): The model to be used for the computation.


        Returns:
            List[MaxScoreSample]: The transformed samples.

        """
        progress = kwargs.get("progress_bar", False)

        for sample in sample_list:
            data = gendered_data[sample.test_case]
            if len(data[0]) > 0:
                macro_f1_score = calculate_f1_score(
                    data[0].to_list(), data[1].to_list(), average="macro", zero_division=0
                )
            else:
                macro_f1_score = 1

            sample.actual_results = MaxScoreOutput(max_score=macro_f1_score)
            sample.state = "done"

            if progress:
                progress.update(1)
        return sample_list


class MinGenderRougeScore(BaseFairness):
    """Subclass of BaseFairness that implements the minimum F1 score.

    Attributes:
        alias_name (str): The name "min_f1" identifying the minimum F1 score.

    Methods:
        transform(data: List[Sample], params: Dict) -> List[MinScoreSample]:
            Transforms the input data into an output based on the minimum F1 score.
    """

    alias_name = [
        "min_gender_rouge1_score",
        "min_gender_rouge2_score",
        "min_gender_rougeL_score",
        "min_gender_rougeLsum_score",
    ]
    supported_tasks = ["question-answering", "summarization"]

    @classmethod
    def transform(
        cls, test: str, data: List[Sample], params: Dict
    ) -> List[MinScoreSample]:
        """Computes the min rouge score for the given data.

        Args:
            test (str): name of the test.
            data (List[Sample]): The input data to be transformed.
            params (Dict): parameters for tests configuration
        Returns:
            List[MinScoreSample]: The transformed data based on the minimum F1 score.
        """
        assert (
            test in cls.alias_name
        ), f"Parameter 'test' should be in: {cls.alias_name}, got '{test}'"

        if isinstance(params["min_score"], dict):
            min_scores = params["min_score"]
        elif isinstance(params["min_score"], float):
            min_scores = {
                "male": params["min_score"],
                "female": params["min_score"],
                "unknown": params["min_score"],
            }

        samples = []
        for key, val in min_scores.items():
            sample = MinScoreSample(
                original=None,
                category="fairness",
                test_type=test,
                test_case=key,
                expected_results=MinScoreOutput(min_score=val),
            )

            samples.append(sample)
        return samples

    @staticmethod
    async def run(
        sample_list: List[MinScoreSample], gendered_data, **kwargs
    ) -> List[MinScoreSample]:
        """Computes the minimum F1 score for the given data.

        Args:
            sample_list (List[MinScoreSample]): The input data to be transformed.
            model (ModelFactory): The model to be used for the computation.

        Returns:
            List[MinScoreSample]: The transformed samples.

        """
        import evaluate

        progress = kwargs.get("progress_bar", False)
        task = kwargs.get("task", None)

        for sample in sample_list:
            data = gendered_data[sample.test_case]
            if len(data[0]) > 0:
                if task == "question-answering" or task == "summarization":
                    em = evaluate.load("rouge")
                    macro_f1_score = em.compute(references=data[0], predictions=data[1])[
                        sample.test_type.split("_")[2]
                    ]
                else:
                    macro_f1_score = calculate_f1_score(
                        [x[0] for x in data[0]], data[1], average="macro", zero_division=0
                    )
            else:
                macro_f1_score = 1

            sample.actual_results = MinScoreOutput(min_score=macro_f1_score)
            sample.state = "done"

            if progress:
                progress.update(1)
        return sample_list


class MaxGenderRougeScore(BaseFairness):
    """Subclass of BaseFairness that implements the rouge score.

    Attributes:
        alias_name (str): The name to be used in config.

    Methods:
        transform(data: List[Sample], params: Dict) -> List[MaxScoreSample]:
            Transforms the input data into an output based on the rouge score.
    """

    alias_name = [
        "max_gender_rouge1_score",
        "max_gender_rouge2_score",
        "max_gender_rougeL_score",
        "max_gender_rougeLsum_score",
    ]
    supported_tasks = ["question-answering", "summarization"]

    @classmethod
    def transform(
        cls, test: str, data: List[Sample], params: Dict
    ) -> List[MaxScoreSample]:
        """Computes the rouge score for the given data.

        Args:
            test (str): name of the test.
            data (List[Sample]): The input data to be transformed.
            params (Dict): parameters for tests configuration
        Returns:
            List[MaxScoreSample]: The transformed data based on the rouge score.
        """
        assert (
            test in cls.alias_name
        ), f"Parameter 'test' should be in: {cls.alias_name}, got '{test}'"

        if isinstance(params["max_score"], dict):
            max_scores = params["max_score"]
        elif isinstance(params["max_score"], float):
            max_scores = {
                "male": params["max_score"],
                "female": params["max_score"],
                "unknown": params["max_score"],
            }

        samples = []
        for key, val in max_scores.items():
            sample = MaxScoreSample(
                original=None,
                category="fairness",
                test_type=test,
                test_case=key,
                expected_results=MaxScoreOutput(max_score=val),
            )

            samples.append(sample)
        return samples

    @staticmethod
    async def run(
        sample_list: List[MaxScoreSample], gendered_data, **kwargs
    ) -> List[MaxScoreSample]:
        """Computes the maximum rouge score for the given data.

        Args:
            sample_list (List[MaxScoreSample]): The input data to be transformed.
            model (ModelFactory): The model to be used for the computation.

        Returns:
            List[MaxScoreSample]: The transformed samples.

        """
        import evaluate

        progress = kwargs.get("progress_bar", False)
        task = kwargs.get("task", None)

        for sample in sample_list:
            data = gendered_data[sample.test_case]
            if len(data[0]) > 0:
                if task == "question-answering" or task == "summarization":
                    em = evaluate.load("rouge")
                    rouge_score = em.compute(references=data[0], predictions=data[1])[
                        sample.test_type.split("_")[2]
                    ]
                else:
                    rouge_score = calculate_f1_score(
                        [x[0] for x in data[0]], data[1], average="macro", zero_division=0
                    )
            else:
                rouge_score = 1

            sample.actual_results = MaxScoreOutput(max_score=rouge_score)
            sample.state = "done"

            if progress:
                progress.update(1)
        return sample_list



import os
import logging

from metaflow import FlowSpec, JSONType, Parameter, step
from transformers import (
    AutoModelForTokenClassification,
    AutoTokenizer,
    Trainer,
    TrainingArguments,
)

from langtest import Harness
from langtest.datahandler.datasource import DataFactory
from langtest.pipelines.utils.data_helpers.ner_dataset import NERDataset
from langtest.pipelines.utils.metrics import compute_ner_metrics


class NEREnd2EndPipeline(FlowSpec):
    """NER pipeline for Huggingface models

    It executes the following workflow in a sequential order:
    - train a model on a given dataset
    - evaluate the model on a given test dataset
    - test the trained model on a set of tests
    - augment the training set based on the tests outcome
    - retrain the model on a the freshly generated augmented training set
    - evaluate the retrained model on the test dataset
    - compare the performance of the two models

    The pipeline can directly be triggered through the CLI via the following one liner:
    ```bash
    python3 langtest/pipelines/transformers_pipelines.py run \
            --model-name="bert-base-uncased" \
            --train-data=tner.csv \
            --eval-data=tner.csv \
            --training-args='{"per_device_train_batch_size": 4, "max_steps": 3}' \
            --feature-col="tokens" \
            --target-col="ner_tags"
    ```
    """

    model_name = Parameter(
        "model-name", help="Name of the pretrained model to load", type=str, required=True
    )
    train_data = Parameter(
        "train-data", help="Path to the train dataset", type=str, required=True
    )
    eval_data = Parameter(
        "eval-data", help="Path to the evaluation dataset", type=str, required=True
    )
    feature_col = Parameter(
        "feature-col",
        help="Name of the feature column to use",
        type=str,
        required=False,
        default="text",
    )
    target_col = Parameter(
        "target-col",
        help="Name of the target column to use",
        type=str,
        required=False,
        default="labels",
    )
    config = Parameter(
        "config", help="Tests configuration", type=JSONType, required=False, default=None
    )
    training_args = Parameter(
        "training-args",
        help="Training arguments to pass to the Trainer",
        type=JSONType,
        required=True,
    )

    @step
    def start(self):
        """Starting step of the flow (required by Metaflow)"""
        self.next(self.setup)

    @step
    def setup(self):
        """Performs all the necessary set up steps"""
        self.task = "ner"
        self.hub = "huggingface"
        self.output_dir = "checkpoints/"

        self.train_datasource = DataFactory(
            file_path={"data_source": self.train_data}, task=self.task
        )
        self.eval_datasource = DataFactory(
            file_path={"data_source": self.eval_data}, task=self.task
        )

        self.next(self.train)

    @step
    def train(self):
        """Performs the training procedure of the model"""
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)

        samples = self.train_datasource.load_raw()
        self.train_dataset = NERDataset(
            tokens=[sample[self.feature_col] for sample in samples],
            labels=[sample[self.target_col] for sample in samples],
            tokenizer=self.tokenizer,
        )

        samples = self.eval_datasource.load_raw()
        self.eval_dataset = NERDataset(
            tokens=[sample[self.feature_col] for sample in samples],
            labels=[sample[self.target_col] for sample in samples],
            tokenizer=self.tokenizer,
        )

        self.model = AutoModelForTokenClassification.from_pretrained(
            self.model_name,
            num_labels=len(self.train_dataset.label_map),
            id2label=self.train_dataset.id2label,
            label2id=self.train_dataset.label_map,
            ignore_mismatched_sizes=True,
        )

        trainer = Trainer(
            model=self.model,
            args=TrainingArguments(output_dir=self.output_dir, **self.training_args),
            train_dataset=self.train_dataset,
            tokenizer=self.tokenizer,
        )
        trainer.train()
        self.model.save_pretrained(self.output_dir)
        self.tokenizer.save_pretrained(self.output_dir)

        self.next(self.evaluate)

    @step
    def evaluate(self):
        """Performs the evaluation procedure on the given test set"""
        trained_model = AutoModelForTokenClassification.from_pretrained(
            self.output_dir,
            num_labels=len(self.train_dataset.label_map),
            id2label=self.train_dataset.id2label,
            label2id=self.train_dataset.label_map,
            ignore_mismatched_sizes=True,
        )
        self.metrics = Trainer(
            model=trained_model,
            eval_dataset=self.eval_dataset,
            compute_metrics=compute_ner_metrics(
                list(self.train_dataset.label_map.keys())
            ),
        ).evaluate()

        self.next(self.test)

    @step
    def test(self):
        """Performs the testing procedure of the model on a set of tests using langtest"""
        self.harness = Harness(
            task=self.task,
            model={"model": self.output_dir, "hub": self.hub},
            data={"data_source": self.train_data},
        )
        if self.config:
            self.harness.configure(self.config)

        _ = self.harness.generate()
        self.harness.save(save_dir="saved_harness")
        _ = self.harness.run()
        self.harness.report(format="dataframe", save_dir="first_report")

        self.next(self.augment)

    @step
    def augment(self):
        """Performs the data augmentation procedure based on langtest"""
        filename = os.path.basename(self.train_data)
        self.path_augmented_file = os.path.join(os.getcwd(), f"augmented_{filename}")
        self.harness.augment(
            training_data={"data_source": self.train_data},
            save_data_path=self.path_augmented_file,
            export_mode="add",
        )

        self.next(self.retrain)

    @step
    def retrain(self):
        """Performs the training procedure using the augmented data created by langtest"""
        self.augmented_train_datasource = DataFactory(
            file_path={"data_source": self.path_augmented_file}, task=self.task
        )
        samples = self.augmented_train_datasource.load_raw()

        self.augmented_train_dataset = NERDataset(
            tokens=[sample["text"] for sample in samples],
            labels=[sample["labels"] for sample in samples],
            tokenizer=self.tokenizer,
        )

        self.model = AutoModelForTokenClassification.from_pretrained(
            self.model_name,
            num_labels=len(self.augmented_train_dataset.label_map),
            id2label=self.augmented_train_dataset.id2label,
            label2id=self.augmented_train_dataset.label_map,
            ignore_mismatched_sizes=True,
        )

        trainer = Trainer(
            model=self.model,
            args=TrainingArguments(
                output_dir=f"augmented_{self.output_dir}", **self.training_args
            ),
            train_dataset=self.augmented_train_dataset,
            eval_dataset=self.eval_dataset,
            tokenizer=self.tokenizer,
        )
        trainer.train()
        self.model.save_pretrained(f"augmented_{self.output_dir}")
        self.tokenizer.save_pretrained(f"augmented_{self.output_dir}")

        self.next(self.reevaluate)

    @step
    def reevaluate(self):
        """Performs the evaluation procedure of the model training on the augmented dataset"""
        newly_trained_model = AutoModelForTokenClassification.from_pretrained(
            f"augmented_{self.output_dir}",
            num_labels=len(self.train_dataset.label_map),
            id2label=self.train_dataset.id2label,
            label2id=self.train_dataset.label_map,
            ignore_mismatched_sizes=True,
        )
        self.augmented_metrics = Trainer(
            model=newly_trained_model,
            eval_dataset=self.eval_dataset,
            compute_metrics=compute_ner_metrics(
                list(self.train_dataset.label_map.keys())
            ),
        ).evaluate()

        self.next(self.compare)

    @step
    def compare(self):
        """Performs the comparison between the two trained models"""
        print("Metrics before augmentation:", self.metrics)
        print("Metrics after augmentation:", self.augmented_metrics)

        self.next(self.end)

    @step
    def end(self):
        """Ending step of the flow (required by Metaflow)"""
        logging.info(f"{self.__class__} successfully ran!")


if __name__ == "__main__":
    NEREnd2EndPipeline()



import os
import random
import re
import string
from abc import ABC, abstractmethod
from collections import defaultdict
from copy import deepcopy as copy
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import yaml

from langtest.datahandler.datasource import DataFactory, HuggingFaceDataset
from langtest.transform import TestFactory
from langtest.transform.utils import create_terminology
from langtest.utils.custom_types import Sample
from langtest.utils.custom_types.output import NEROutput
from langtest.utils.custom_types.predictions import NERPrediction, SequenceLabel
from langtest.utils.custom_types.sample import NERSample


class BaseAugmentaion(ABC):
    """Abstract base class for data augmentation techniques.

    Methods:
        fix: Abstract method that should be implemented by child classes.
             This method should perform the data augmentation operation.
    """

    @abstractmethod
    def fix(self, *args, **kwargs):
        """Abstract method that should be implemented by child classes.

        This method should perform the data augmentation operation.

        Returns:
            NotImplementedError: Raised if the method is not implemented by child classes.
        """
        return NotImplementedError


class AugmentRobustness(BaseAugmentaion):
    """A class for performing a specified task with historical results.

    Attributes:
        task (str): A string indicating the task being performed.
        config (dict): A dictionary containing configuration parameters for the task.
        h_report (pandas.DataFrame): A DataFrame containing a report of historical results for the task.
        max_prop (float): The maximum proportion of improvement that can be suggested by the class methods.
                        Defaults to 0.5.

    Methods:
        __init__(self, task, h_report, config, max_prop=0.5) -> None:
            Initializes an instance of MyClass with the specified parameters.

        fix(self) -> List[Sample]:
            .

        suggestions(self, prop) -> pandas.DataFrame:
            Calculates suggestions for improving test performance based on a given report.
    """

    def __init__(
        self,
        task: str,
        h_report: "pd.DataFrame",
        config: Dict,
        custom_proportions: Union[Dict, List] = None,
        max_prop=0.5,
    ) -> None:
        """Initializes an instance of MyClass with the specified parameters.

        Args:
            task (str): A string indicating the task being performed.
            h_report (pandas.DataFrame): A DataFrame containing a report of historical results for the task.
            config (dict): A dictionary containing configuration parameters for the task.
            custom_proportions
            max_prop (float): The maximum proportion of improvement that can be suggested by the class methods.
                              Defaults to 0.5.

        Returns:
            None

        """
        super().__init__()
        self.task = task
        self.config = config
        self.h_report = h_report
        self.max_prop = max_prop
        self.custom_proportions = custom_proportions

        if isinstance(self.config, str):
            with open(self.config) as fread:
                self.config = yaml.safe_load(fread)

    def fix(
        self,
        training_data: dict,
        output_path: str,
        export_mode: str = "add",
    ):
        """Applies perturbations to the input data based on the recommendations from harness reports.

        Args:
            training_data (dict): A dictionary containing the input data for augmentation.
            output_path (str): The path to save the augmented data file.
            export_mode (str, optional): Determines how the samples are modified or exported.
                                        - 'inplace': Modifies the list of samples in place.
                                        - 'add': Adds new samples to the input data.
                                        - 'transformed': Exports only the transformed data, excluding untransformed samples.
                                        Defaults to 'add'.

        Returns:
            List[Dict[str, Any]]: A list of augmented data samples.
        """

        if "source" in training_data and training_data["source"] == "huggingface":
            self.df = HuggingFaceDataset(training_data["data_source"], self.task)
            data = self.df.load_data(
                feature_column=training_data.get("feature_column", "text"),
                target_column=training_data.get("target_column", "label"),
                split=training_data.get("split", "test"),
                subset=training_data.get("subset", None),
            )
        else:
            self.df = DataFactory(training_data, self.task)
            data = self.df.load()
        TestFactory.is_augment = True
        supported_tests = TestFactory.test_scenarios()
        suggest: pd.DataFrame = self.suggestions(self.h_report)
        sum_propotion = suggest["proportion_increase"].sum()
        if suggest.shape[0] <= 0 or suggest.empty:
            print("All tests have passed. Augmentation will not be applied in this case.")
            return None

        self.config = self._parameters_overrides(self.config, data)

        final_aug_data = []
        hash_map = {k: v for k, v in enumerate(data)}
        transformed_data = []
        for proportion in suggest.iterrows():
            cat = proportion[-1]["category"].lower()
            if cat not in ["robustness", "bias"]:
                continue
            test = proportion[-1]["test_type"].lower()
            test_type = {cat: {test: self.config.get("tests").get(cat).get(test)}}
            if proportion[-1]["test_type"] in supported_tests[cat]:
                sample_length = (
                    len(data)
                    * self.max_prop
                    * (proportion[-1]["proportion_increase"] / sum_propotion)
                )
                if export_mode in ("inplace"):
                    sample_indices = random.sample(
                        range(0, len(data)), int(sample_length)
                    )
                    for each in sample_indices:
                        if test == "swap_entities":
                            test_type["robustness"]["swap_entities"]["parameters"][
                                "labels"
                            ] = [self.label[each]]
                        res = TestFactory.transform(
                            self.task, [hash_map[each]], test_type
                        )
                        if len(res) == 0:
                            continue
                        hash_map[each] = res[0]
                else:
                    if test == "swap_entities":
                        sample_data = data[: int(sample_length)]
                        test_type["robustness"]["swap_entities"]["parameters"][
                            "labels"
                        ] = test_type["robustness"]["swap_entities"]["parameters"][
                            "labels"
                        ][
                            : int(sample_length)
                        ]
                    else:
                        sample_data = random.choices(data, k=int(sample_length))
                    aug_data = TestFactory.transform(self.task, sample_data, test_type)
                    final_aug_data.extend(aug_data)

                    if export_mode == "transformed":
                        transformed_data.extend(aug_data)
        if "." not in training_data["data_source"]:
            if export_mode == "inplace":
                final_aug_data = list(hash_map.values())
                self.df.export_data(final_aug_data, output_path)
            elif export_mode == "transformed":
                self.df.export_data(transformed_data, output_path)
            else:
                data.extend(final_aug_data)
                self.df.export_data(data, output_path)

            TestFactory.is_augment = False
            return final_aug_data

        else:
            if export_mode == "inplace":
                final_aug_data = list(hash_map.values())
                self.df.export(final_aug_data, output_path)
            elif export_mode == "transformed":
                self.df.export(transformed_data, output_path)
            else:
                data.extend(final_aug_data)
                self.df.export(data, output_path)

            TestFactory.is_augment = False
            return final_aug_data

    def suggestions(self, report: "pd.DataFrame") -> "pd.DataFrame":
        """Calculates suggestions for improving test performance based on a given report.

        Args:
            report (pandas.DataFrame): A DataFrame containing test results by category and test type,
                                        including pass rates and minimum pass rates.

        Returns:
            pandas.DataFrame: A DataFrame containing the following columns for each suggestion:
                                - category: the test category
                                - test_type: the type of test
                                - ratio: the pass rate divided by the minimum pass rate for the test
                                - proportion_increase: a proportion indicating how much the pass rate
                                                    should increase to reach the minimum pass rate
        """
        report["ratio"] = report["pass_rate"] / report["minimum_pass_rate"]

        if self.custom_proportions and isinstance(self.custom_proportions, dict):
            report["proportion_increase"] = report["test_type"].map(
                self.custom_proportions
            )
        elif self.custom_proportions and isinstance(self.custom_proportions, list):
            report["proportion_increase"] = report["ratio"].apply(self._proportion_values)
            report = report[report["test_type"].isin(self.custom_proportions)]
        else:
            report["proportion_increase"] = report["ratio"].apply(self._proportion_values)

        report = report.dropna(subset=["proportion_increase"])[
            ["category", "test_type", "ratio", "proportion_increase"]
        ]
        return report

    @staticmethod
    def _proportion_values(x: float) -> Optional[float]:
        """Calculates a proportion indicating how much a pass rate should increase to reach a minimum pass rate.

        Args:
            x (float): The ratio of the pass rate to the minimum pass rate for a given test.

        Returns:
            float: A proportion indicating how much the pass rate should increase to reach the minimum pass rate.
                If the pass rate is greater than or equal to the minimum pass rate, returns None.
                If the pass rate is between 0.9 and 1.0 times the minimum pass rate, returns 0.05.
                If the pass rate is between 0.8 and 0.9 times the minimum pass rate, returns 0.1.
                If the pass rate is between 0.7 and 0.8 times the minimum pass rate, returns 0.2.
                If the pass rate is less than 0.7 times the minimum pass rate, returns 0.3.

        """
        if x >= 1:
            return None
        elif x > 0.9:
            return 0.05
        elif x > 0.8:
            return 0.1
        elif x > 0.7:
            return 0.2
        else:
            return 0.3

    def _parameters_overrides(self, config: dict, data_handler: List[Sample]) -> dict:
        tests = config.get("tests", {}).get("robustness", {})
        if "swap_entities" in config.get("tests", {}).get("robustness", {}):
            df = pd.DataFrame(
                {
                    "text": [sample.original for sample in data_handler],
                    "label": [
                        [i.entity for i in sample.expected_results.predictions]
                        for sample in data_handler
                    ],
                }
            )
            params = tests["swap_entities"]
            params["parameters"] = {}
            params["parameters"]["terminology"] = create_terminology(df)
            params["parameters"]["labels"] = df.label.tolist()
            self.label = (
                self.config.get("tests")
                .get("robustness")
                .get("swap_entities")
                .get("parameters")
                .get("labels")
            )
        return config


class TemplaticAugment(BaseAugmentaion):
    """This class is used for templatic augmentation. It is a subclass of the BaseAugmentation class.

    Attributes:
        __templates:
            A string or a list of strings or samples that represents the templates for the augmentation.
        __task:
            The task for which the augmentation is being performed.

    Methods:
        __init__(self, templates: Union[str, List[str]], task: str):
            Initializes the TemplaticAugment class.
        fix(self, training_data: str, output_path: str, *args, **kwargs):
            Performs the templatic augmentation and exports the results to a specified path.
    """

    def __init__(self, templates: Union[str, List[str]], task: str) -> None:
        """This constructor for the TemplaticAugment class.

        Args:
            templates (Union[str, List[str]]): The templates to be used for the augmentation.
            task (str): The task for which the augmentation is being performed.
        """
        self.__templates: Union[str, List[str], List[Sample]] = templates
        self.__task = task

        if isinstance(self.__templates, str) and os.path.exists(self.__templates):
            self.__templates = DataFactory(self.__templates, self.__task).load()
        elif isinstance(self.__templates, str):
            self.__templates = [self.str_to_sample(self.__templates)]
        elif isinstance(self.__templates, list) and isinstance(self.__templates[0], str):
            self.__templates = [self.str_to_sample(i) for i in self.__templates]

    def fix(
        self,
        training_data: Dict[str, Any],
        output_path: str,
        max_num: int = None,
        *args,
        **kwargs,
    ) -> bool:
        """This method is used to perform the templatic augmentation.

        It takes the input data, performs the augmentation and then saves the augmented data to the output path.

        Args:
            training_data (dict): A dictionary containing the input data for augmentation.
            output_path (str): The path where the augmented data will be saved.
            max_num (int): Maximum number of new samples to generate
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            bool: Returns True upon successful completion of the method.
        """
        df = DataFactory(training_data, self.__task)
        data = df.load()
        new_data = []
        self.__search_results = self.search_sample_results(data)
        if not max_num:
            max_num = max(len(i) for i in self.__search_results.values())

        for template in self.__templates:
            for _ in range(max_num):
                new_sample = self.new_sample(template)
                if new_sample:
                    new_data.append(new_sample)

        df.export(new_data, output_path)
        return True

    @staticmethod
    def search_sample_results(
        samples: List[Sample],
    ) -> Dict[str, List[Union[NERPrediction, SequenceLabel]]]:
        """This method is used to search the results of the samples for the entities in the templates.

        Args:
            samples (List[Sample]): The samples for which the results are to be searched.

        Returns:
            Dict[str, List[Union[NERPrediction, SequenceLabel]]]: A dictionary containing the search results.
        """
        results_dict = defaultdict(list)
        for sample in samples:
            chunk = []
            ent_name = ""
            for result in sample.expected_results.predictions:
                ent = result.entity.split("-")[-1]
                if ent != "O" and ent_name == "":
                    ent_name = ent
                if result.entity.endswith(ent_name) and ent != "O":
                    result.doc_id = 0
                    result.doc_name = ""
                    chunk.append(result)
                elif len(chunk) > 0:
                    results_dict[ent_name].append(tuple(chunk))
                    ent_name = ""
                    chunk = []

            if chunk:
                results_dict[ent_name].append(tuple(chunk))
        return results_dict

    @staticmethod
    def extract_variable_names(f_string: str) -> List[str]:
        """This method is used to extract the variable names from the templates.

        Args:
            f_string (str): The template string.

        Returns:
            List[str]: A list of variable names.
        """
        pattern = r"{([^{}]*)}"
        matches = re.findall(pattern, f_string)
        variable_names = [match.strip() for match in matches]
        return variable_names

    def new_sample(self, template: Sample):
        """This method is used to generate a new sample from a template.

        Args:
            template (Sample): The template from which the new sample is to be generated.

        Returns:
            Sample: The new sample generated from the template.
        """
        template = copy(template)
        matches = re.finditer(r"{([^{}]*)}", template.original)
        cursor = 0
        other_predictions = []
        if matches:
            for match in matches:
                entity = match.group(1)
                if entity in self.__search_results:
                    prediction = random.choice(self.__search_results[entity])
                    word = " ".join(
                        i.span.word for i in prediction if isinstance(i, NERPrediction)
                    )

                    template.original = template.original.replace(
                        "{" + entity + "}", word, 1
                    )
                    for result in template.expected_results.predictions[cursor:]:
                        if prediction[0].entity.endswith(result.entity):
                            for each_prediction in prediction:
                                if isinstance(each_prediction, NERPrediction):
                                    each_prediction.chunk_tag = "-X-"
                                    each_prediction.pos_tag = "-X-"
                            other_predictions.extend(prediction)
                            cursor += 1
                            break
                        else:
                            if "{" in result.span.word and "}" in result.span.word:
                                continue
                            other_predictions.append(result)
                            cursor += 1
                else:
                    continue
            template.expected_results.predictions = (
                other_predictions + template.expected_results.predictions[cursor:]
            )
            return template
        else:
            return None

    def str_to_sample(self, template: str):
        """This method is used to convert a template string to a Sample object.

        Args:
            template (str): The template string to be converted.

        Returns:
            Sample: The Sample object generated from the template string.
        """
        if self.__task == "ner":
            template = self.add_spaces_around_punctuation(template)
            sample = NERSample()
            sample.original = template
            words = template.split()
            predictions = []
            cursor = 0
            for word in words:
                if "{" in word and "}" in word:
                    entity = word.replace("{", "").replace("}", "")
                else:
                    entity = "O"
                predictions.append(
                    NERPrediction.from_span(
                        entity,
                        word,
                        cursor,
                        cursor + len(word),
                        pos_tag="-X-",
                        chunk_tag="-X-",
                        doc_id=0,
                        doc_name="",
                    )
                )
                cursor += len(word) + 1
            sample.expected_results = NEROutput(predictions=predictions)

        elif self.__task == "text-classification":
            raise NotImplementedError

        return sample

    @property
    def templates(self):
        """Templates getter"""
        return self.__templates

    @templates.setter
    def templates(self, templates: Union[str, List[str]]):
        self.__init__(templates, self.__task)

    @property
    def task(self):
        """Task getter"""
        return self.__task

    @task.setter
    def task(self, task: str):
        self.__task = task

    @staticmethod
    def add_spaces_around_punctuation(text: str):
        """This method is used to add spaces around punctuation in a string.

        Args:
            text (str): The string to which spaces are to be added.

        Returns:
            str: The string with spaces added around punctuation.
        """
        for punct in string.punctuation:
            if punct not in ["{", "}", "_"]:
                if punct == ".":
                    # To prevent spaces being added around decimal points
                    text = re.sub(r"(\d)\.(\d)", r"\1[DOT]\2", text)

                text = text.replace(punct, f" {punct} ")

                if punct == ".":
                    # Putting back the decimal points to original state
                    text = text.replace("[DOT]", ".")

        # Removing extra spaces
        text = re.sub(r"\s+", " ", text).strip()

        return text



import asyncio
import random
import re
import string
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Dict, List, Optional, Tuple, Union

from langtest.modelhandler.modelhandler import ModelFactory
from langtest.utils.custom_types import SequenceClassificationSample
from .constants import (
    CONTRACTION_MAP,
    Slang_Adjectives,
    Slang_Adverbs,
    Slang_Nouns,
    TYPO_FREQUENCY,
    abbreviation_dict,
    dyslexia_map,
    ocr_typo_dict,
    adjective_synonym_dict,
    adjective_antonym_dict,
)
from ..utils.SoundsLikeFunctions import Search
from ..utils.custom_types import Sample, Span, Transformation
from ..utils.number_to_word import ConvertNumberToWord
from collections import defaultdict


inverted_ocr_typo_dict = defaultdict(list)
for k, v in ocr_typo_dict.items():
    inverted_ocr_typo_dict[v].append(k)


class BaseRobustness(ABC):
    """Abstract base class for implementing robustness measures.

    Attributes:
        alias_name (str): A name or list of names that identify the robustness measure.

    Methods:
        transform(data: List[Sample]) -> Any: Transforms the input data into an output based on the implemented robustness measure.
    """

    alias_name = None
    supported_tasks = [
        "ner",
        "text-classification",
        "question-answering",
        "summarization",
        "translation",
    ]

    @staticmethod
    @abstractmethod
    def transform(sample_list: List[Sample]) -> List[Sample]:
        """Abstract method that implements the robustness measure.

        Args:
            sample_list (List[Sample]): The input data to be transformed.

        Returns:
            Any: The transformed data based on the implemented robustness measure.
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    async def run(
        sample_list: List[Sample], model: ModelFactory, **kwargs
    ) -> List[Sample]:
        """Abstract method that implements the robustness measure.

        Args:
            sample_list (List[Sample]): The input data to be transformed.
            model (ModelFactory): The model to be used for evaluation.
            **kwargs: Additional arguments to be passed to the robustness measure.

        Returns:
            List[Sample]: The transformed data based on the implemented robustness measure.

        """
        progress = kwargs.get("progress_bar", False)
        for sample in sample_list:
            if sample.state != "done":
                if hasattr(sample, "run"):
                    sample_status = sample.run(model, **kwargs)
                    if sample_status:
                        sample.state = "done"
                else:
                    sample.expected_results = model(sample.original)
                    sample.actual_results = model(sample.test_case)
                    sample.state = "done"
            if progress:
                progress.update(1)
        return sample_list

    @classmethod
    async def async_run(cls, sample_list: List[Sample], model: ModelFactory, **kwargs):
        """Creates a task to run the robustness measure.

        Args:
            sample_list (List[Sample]): The input data to be transformed.
            model (ModelFactory): The model to be used for evaluation.
            **kwargs: Additional arguments to be passed to the robustness measure.

        Returns:
            asyncio.Task: The task that runs the robustness measure.

        """
        created_task = asyncio.create_task(cls.run(sample_list, model, **kwargs))
        return created_task


class UpperCase(BaseRobustness):
    """A class for transforming text samples to uppercase."""

    alias_name = "uppercase"

    @staticmethod
    def transform(sample_list: List[Sample], prob: Optional[float] = 1.0) -> List[Sample]:
        """Transform the text samples in the given sample list to uppercase.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of transforming each word to uppercase.
                Defaults to 1.0, which means all words will be transformed.

        Returns:
            List[Sample]: The transformed sample list with text samples in uppercase.
        """
        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                words = sample.split()
                num_transform_words = int(prob * len(words))
                transformed_indices = random.sample(
                    range(len(words)), num_transform_words
                )
                transformed_words = [
                    words[i].upper() if i in transformed_indices else words[i]
                    for i in range(len(words))
                ]
                sample_list[idx] = " ".join(transformed_words)
            else:
                words = sample.original.split()
                num_transform_words = int(prob * len(words))
                transformed_indices = random.sample(
                    range(len(words)), num_transform_words
                )
                transformed_words = [
                    words[i].upper() if i in transformed_indices else words[i]
                    for i in range(len(words))
                ]
                sample.test_case = " ".join(transformed_words)
                sample.category = "robustness"
        return sample_list


class LowerCase(BaseRobustness):
    """A class for transforming text samples to lowercase."""

    alias_name = "lowercase"

    @staticmethod
    def transform(sample_list: List[Sample], prob: Optional[float] = 1.0) -> List[Sample]:
        """Transform the text samples in the given sample list to lowercase.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of transforming each word to lowercase.
                                    Defaults to 1.0, which means all words will be transformed.

        Returns:
            List[Sample]: The transformed sample list with text samples in lowercase.
        """
        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                words = sample.split()
                num_transform_words = int(prob * len(words))
                transformed_indices = random.sample(
                    range(len(words)), num_transform_words
                )
                transformed_words = [
                    words[i].lower() if i in transformed_indices else words[i]
                    for i in range(len(words))
                ]
                sample_list[idx] = " ".join(transformed_words)
            else:
                words = sample.original.split()
                num_transform_words = int(prob * len(words))
                transformed_indices = random.sample(
                    range(len(words)), num_transform_words
                )
                transformed_words = [
                    words[i].lower() if i in transformed_indices else words[i]
                    for i in range(len(words))
                ]
                sample.test_case = " ".join(transformed_words)
                sample.category = "robustness"
        return sample_list


class TitleCase(BaseRobustness):
    """A class for transforming text samples to titlecase."""

    alias_name = "titlecase"

    @staticmethod
    def transform(sample_list: List[Sample], prob: Optional[float] = 1.0) -> List[Sample]:
        """Transform the text samples in the given sample list to title case.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of transforming each word to title case.
                                    Defaults to 1.0, which means all words will be transformed.

        Returns:
            List[Sample]: The transformed sample list with text samples in title case.
        """
        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                words = sample.split()
                num_transform_words = int(prob * len(words))
                transformed_indices = random.sample(
                    range(len(words)), num_transform_words
                )
                transformed_words = [
                    words[i].title() if i in transformed_indices else words[i]
                    for i in range(len(words))
                ]
                sample_list[idx] = " ".join(transformed_words)
            else:
                words = sample.original.split()
                num_transform_words = int(prob * len(words))
                transformed_indices = random.sample(
                    range(len(words)), num_transform_words
                )
                transformed_words = [
                    words[i].title() if i in transformed_indices else words[i]
                    for i in range(len(words))
                ]
                sample.test_case = " ".join(transformed_words)
                sample.category = "robustness"
        return sample_list


class AddPunctuation(BaseRobustness):
    """A class for adding punctuation to text samples."""

    alias_name = "add_punctuation"

    @staticmethod
    def transform(
        sample_list: List[Sample],
        prob: Optional[float] = 1.0,
        whitelist: Optional[List[str]] = None,
        count: int = 1,
    ) -> List[Sample]:
        """Add punctuation at the end of the string, if there is punctuation at the end skip it

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of adding punctuation to each sample.
                                    Defaults to 1.0, which means all words will be transformed.
            whitelist (Optional[List[str]]): A list of punctuation characters to consider when adding punctuation.
                If None, the default whitelist ['!', '?', ',', '.', '-', ':', ';'] will be used. Defaults to None.
            count: Number of variations to create.

        Returns:
            List[Sample]: The transformed sample list with added punctuation.
        """
        if whitelist is None:
            whitelist = ["!", "?", ",", ".", "-", ":", ";"]

        def check_whitelist(text, whitelist):
            if text[-1] not in whitelist:
                chosen_punc = random.choice(whitelist)
                return text + chosen_punc
            else:
                return text

        perturbed_samples = []

        for s in sample_list:
            sample = deepcopy(s)
            for i in range(count):
                if isinstance(sample, str):
                    if random.random() < prob:
                        perturbed_samples.append(check_whitelist(sample, whitelist))
                else:
                    if sample.original[-1] not in whitelist and (random.random() < prob):
                        chosen_punc = random.choice(whitelist)
                        sample.test_case = sample.original + chosen_punc
                        if sample.task in ("ner", "text-classification"):
                            sample.transformations = [
                                Transformation(
                                    original_span=Span(
                                        start=len(sample.original),
                                        end=len(sample.original),
                                        word="",
                                    ),
                                    new_span=Span(
                                        start=len(sample.original),
                                        end=len(sample.test_case),
                                        word=chosen_punc,
                                    ),
                                    ignore=True,
                                )
                            ]
                    else:
                        sample.test_case = sample.original
                    sample.category = "robustness"
                    perturbed_samples.append(sample)
        return perturbed_samples


class StripPunctuation(BaseRobustness):
    """A class for stripping punctuation to text samples."""

    alias_name = "strip_punctuation"

    @staticmethod
    def transform(
        sample_list: List[Sample],
        prob: Optional[float] = 1.0,
        whitelist: Optional[List[str]] = None,
    ) -> List[Sample]:
        """Strip punctuation from the text samples in the given sample list.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of stripping punctuation from each sample.
                                    Defaults to 1.0, which means all words will be transformed.
            whitelist (Optional[List[str]]): A list of punctuation characters to consider when stripping punctuation.
                If None, the default whitelist ['!', '?', ',', '.', '-', ':', ';'] will be used. Defaults to None.

        Returns:
            List[Sample]: The transformed sample list with punctuation stripped.
        """
        if whitelist is None:
            whitelist = ["!", "?", ",", ".", "-", ":", ";"]

        def check_whitelist(text, whitelist):
            if text[-1] in whitelist:
                return text[:-1]
            else:
                return text

        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                if random.random() < prob:
                    sample_list[idx] = check_whitelist(sample, whitelist)
            else:
                if sample.original[-1] in whitelist and (random.random() < prob):
                    sample.test_case = sample.original[:-1]
                    if sample.task in ("ner", "text-classification"):
                        sample.transformations = [
                            Transformation(
                                original_span=Span(
                                    start=len(sample.original) - 1,
                                    end=len(sample.original),
                                    word=sample.original[-1:],
                                ),
                                new_span=Span(
                                    start=len(sample.test_case),
                                    end=len(sample.test_case),
                                    word="",
                                ),
                                ignore=True,
                            )
                        ]
                else:
                    sample.test_case = sample.original
                sample.category = "robustness"
        return sample_list


class AddTypo(BaseRobustness):
    """A class for adding typos to text samples."""

    alias_name = "add_typo"

    @staticmethod
    def transform(
        sample_list: List[Sample], prob: Optional[float] = 1.0, count: int = 1
    ) -> List[Sample]:
        """Add typos to the text samples in the given sample list.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of adding a typo to each sample.
                                    Defaults to 1.0, which means all words will be transformed.

            count: Number of variations to create.

        Returns:
            List[Sample]: The transformed sample list with added typos.
        """

        def keyboard_typo(string):
            if prob is not None and random.random() >= prob:
                return string

            if len(string) < 5:
                return string
            string = list(string)
            if random.random() > 0.1:
                idx_list = list(range(len(TYPO_FREQUENCY)))
                char_list = list(TYPO_FREQUENCY.keys())

                counter, idx = 0, -1
                while counter < 10 and idx == -1:
                    idx = random.randint(0, len(string) - 1)
                    char = string[idx]
                    if TYPO_FREQUENCY.get(char.lower(), None):
                        char_frequency = TYPO_FREQUENCY[char.lower()]

                        if sum(char_frequency) > 0:
                            chosen_char = random.choices(idx_list, weights=char_frequency)
                            difference = ord(char.lower()) - ord(
                                char_list[chosen_char[0]]
                            )
                            char = chr(ord(char) - difference)
                            string[idx] = char
                    else:
                        idx = -1
                        counter += 1
            else:
                string = list(string)
                swap_idx = random.randint(0, len(string) - 2)
                tmp = string[swap_idx]
                string[swap_idx] = string[swap_idx + 1]
                string[swap_idx + 1] = tmp
            return "".join(string)

        perturbed_samples = []
        for sample in sample_list:
            for i in range(count):
                if isinstance(sample, str):
                    perturbed_samples.append(keyboard_typo(sample))
                else:
                    s = deepcopy(sample)
                    s.category = "robustness"
                    s.test_case = keyboard_typo(sample.original)
                    perturbed_samples.append(s)

        return perturbed_samples


class SwapEntities(BaseRobustness):
    """A class for swapping entities in text samples."""

    alias_name = "swap_entities"
    supported_tasks = ["ner"]

    @staticmethod
    def transform(
        sample_list: List[Sample],
        prob: Optional[float] = 1.0,
        labels: List[List[str]] = None,
        terminology: Dict[str, List[str]] = None,
        count: int = 1,
    ) -> List[Sample]:
        """Swap entities in the text samples of the given sample list.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of swapping entities in each sample.
                                    Defaults to 1.0, which means all eligible samples will have their entities swapped.
            labels (List[List[str]]):  Corresponding labels to make changes according to sentences.
            terminology (Dict[str, List[str]]): Dictionary of entities and corresponding list of words.

        Returns:
            List[Sample]: The transformed sample list with entities swapped.
        """
        if terminology is None:
            raise ValueError(
                "In order to generate test cases for swap_entities, terminology should be passed!"
            )

        if labels is None:
            raise ValueError(
                "In order to generate test cases for swap_entities, labels should be passed!"
            )

        assert len(sample_list) == len(
            labels
        ), "'labels' and 'sample_list' must have same lengths."

        perturbed_samples = []
        for s, sample_labels in zip(sample_list, labels):
            for i in range(count):
                sample = deepcopy(s)
                sample.category = "robustness"
                if all([label == "O" for label in sample_labels]):
                    sample.test_case = sample.original
                    break

                sent_tokens = sample.original.split(" ")

                ent_start_pos = [1 if label[0] == "B" else 0 for label in sample_labels]
                ent_idx = [i for i, value in enumerate(ent_start_pos) if value == 1]

                if not ent_idx:
                    sample.test_case = sample.original
                    break

                replace_idx = random.choice(ent_idx)
                ent_type = sample_labels[replace_idx][2:]
                replace_idxs = [replace_idx]
                if replace_idx < len(sample_labels) - 1:
                    for i, label in enumerate(sample_labels[replace_idx + 1 :]):
                        if label == f"I-{ent_type}":
                            replace_idxs.append(i + replace_idx + 1)
                        else:
                            break

                replace_token = sent_tokens[replace_idx : replace_idx + len(replace_idxs)]
                token_length = len(replace_token)
                replace_token = " ".join(replace_token)
                filtered_ents = [
                    ent
                    for ent in terminology[ent_type]
                    if len(ent.split(" ")) == token_length
                ]
                if not filtered_ents:
                    chosen_ent = random.choice(terminology[ent_type])
                else:
                    chosen_ent = random.choice(filtered_ents)

                if random.random() < prob:
                    replace_token_pos = re.search(
                        re.escape(replace_token), sample.original
                    )
                    sample.test_case = sample.original.replace(replace_token, chosen_ent)
                    if (
                        sample.task in ("ner", "text-classification")
                        and replace_token_pos
                    ):
                        sample.transformations = [
                            Transformation(
                                original_span=Span(
                                    start=replace_token_pos.start(),
                                    end=replace_token_pos.end(),
                                    word=replace_token,
                                ),
                                new_span=Span(
                                    start=replace_token_pos.start(),
                                    end=replace_token_pos.start() + len(chosen_ent),
                                    word=chosen_ent,
                                ),
                                ignore=False,
                            )
                        ]
                else:
                    sample.test_case = sample.original
                perturbed_samples.append(sample)

        return perturbed_samples


class ConvertAccent(BaseRobustness):
    """A class for converting accents in text samples."""

    alias_name = ["american_to_british", "british_to_american"]

    @staticmethod
    def transform(
        sample_list: List[Sample],
        prob: Optional[float] = 1.0,
        accent_map: Dict[str, str] = None,
    ) -> List[Sample]:
        """Converts accents in the input sentences using a conversion dictionary.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of applying accent conversion to each sample.
                                    Defaults to 1.0, which means all samples will have accents converted.
            accent_map (Dict[str, str]): A dictionary with conversion terms mapping source accents to target accents.

        Returns:
            List[Sample]: The transformed sample list with accents converted.
        """

        def convert_accent(string: str, accent_map: Dict[str, str]) -> str:
            tokens = set(string.split(" "))
            replaced_string = string
            transformations = []

            for i, token in enumerate(tokens):
                if random.random() < prob:
                    new_token = accent_map.get(token.lower(), token)
                    if new_token != token:
                        diff_len = len(new_token) - len(token)
                        nb_occurrences = len(re.findall(token, replaced_string))

                        for c in range(nb_occurrences):
                            span = re.search(token, replaced_string)
                            replaced_string = re.sub(
                                token, new_token, replaced_string, count=1
                            )

                            transformations.append(
                                Transformation(
                                    original_span=Span(
                                        start=span.start(), end=span.end(), word=token
                                    ),
                                    new_span=Span(
                                        start=span.start(),
                                        end=span.end() + diff_len,
                                        word=new_token,
                                    ),
                                    ignore=False,
                                )
                            )
            return replaced_string, transformations

        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                sample_list[idx], _ = convert_accent(sample, accent_map)
            else:
                if sample.task in ("ner", "text-classification"):
                    sample.test_case, sample.transformations = convert_accent(
                        sample.original, accent_map
                    )
                else:
                    sample.test_case, _ = convert_accent(sample.original, accent_map)
                sample.category = "robustness"

        return sample_list


class AddContext(BaseRobustness):
    """A class for adding context to text samples."""

    alias_name = "add_context"

    @staticmethod
    def transform(
        sample_list: List[Sample],
        prob: Optional[float] = 1.0,
        starting_context: Optional[List[str]] = None,
        ending_context: Optional[List[str]] = None,
        strategy: str = None,
        count: int = 1,
    ) -> List[Sample]:
        """Adds context to the input sentences.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of adding context to each sample.
                                     Defaults to 1.0, which means all samples will have context added.
            starting_context (Optional[List[str]]): A list of terms (context) to be added at the start of sentences.
            ending_context (Optional[List[str]]): A list of terms (context) to be added at the end of sentences.
            strategy (str): Config method to adjust where the context tokens are added. Options: 'start', 'end', or 'combined'.
            count: Number of variations to create.

        Returns:
            List[Sample]: The transformed sample list with context added.
        """

        def context(text, strategy):
            possible_methods = ["start", "end", "combined"]
            if strategy is None:
                strategy = random.choice(possible_methods)
            elif strategy not in possible_methods:
                raise ValueError(
                    f"Add context strategy must be one of 'start', 'end', 'combined'. Cannot be {strategy}."
                )

            transformations = []

            if strategy == "start" or strategy == "combined":
                if random.random() < prob:
                    add_tokens = random.choice(starting_context)
                    add_string = (
                        " ".join(add_tokens)
                        if isinstance(add_tokens, list)
                        else add_tokens
                    )
                    if text != "-":
                        text = add_string + " " + text
                        transformations.append(
                            Transformation(
                                original_span=Span(start=0, end=0, word=""),
                                new_span=Span(
                                    start=0, end=len(add_string) + 1, word=add_string
                                ),
                                ignore=True,
                            )
                        )

            if strategy == "end" or strategy == "combined":
                if random.random() < prob:
                    add_tokens = random.choice(ending_context)
                    add_string = (
                        " ".join(add_tokens)
                        if isinstance(add_tokens, list)
                        else add_tokens
                    )
                    if text != "-":
                        text = text + " " + add_string
                        transformations.append(
                            Transformation(
                                original_span=Span(
                                    start=len(text) - 1, end=len(text), word=""
                                ),
                                new_span=Span(
                                    start=len(text),
                                    end=len(text) + len(add_string) + 1,
                                    word=add_string,
                                ),
                                ignore=True,
                            )
                        )

            return text, transformations

        perturbed_samples = []
        for s in sample_list:
            for i in range(count):
                sample = deepcopy(s)
                if isinstance(sample, str):
                    sample, _ = context(sample, strategy)
                else:
                    if sample.task in ("ner", "text-classification"):
                        sample.test_case, sample.transformations = context(
                            sample.original, strategy
                        )
                    else:
                        sample.test_case, _ = context(sample.original, strategy)

                    sample.category = "robustness"
                perturbed_samples.append(sample)
        return perturbed_samples


class AddContraction(BaseRobustness):
    """A class for adding contractions to text samples."""

    alias_name = "add_contraction"

    @staticmethod
    def transform(sample_list: List[Sample], prob: Optional[float] = 1.0) -> List[Sample]:
        """Adds contractions to the input sentences.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of adding contractions to each sample.
                                    Defaults to 1.0, which means all samples will have context added.

        Returns:
            List[Sample]: The transformed sample list with contractions added.
        """

        def custom_replace(match):
            """Regex replace for contraction."""
            token = match.group(0)
            contracted_token = CONTRACTION_MAP.get(
                token, CONTRACTION_MAP.get(token.lower())
            )

            is_upper_case = token[0]
            expanded_contraction = is_upper_case + contracted_token[1:]
            return expanded_contraction

        def search_contraction(text):
            replaced_string = text
            for contraction in CONTRACTION_MAP:
                search = re.search(contraction, text, flags=re.IGNORECASE | re.DOTALL)
                if search and (random.random() < prob):
                    replaced_string = re.sub(
                        contraction,
                        custom_replace,
                        replaced_string,
                        flags=re.IGNORECASE | re.DOTALL,
                    )

            return replaced_string

        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                sample_list[idx] = search_contraction(sample)
            else:
                replaced_string = sample.original
                transformations = []

                for contraction in CONTRACTION_MAP:
                    search = re.search(
                        contraction, sample.original, flags=re.IGNORECASE | re.DOTALL
                    )
                    if search and (random.random() < prob):
                        new_string = CONTRACTION_MAP.get(search.group(), search.group())

                        diff_len = len(new_string) - len(search.group())
                        replaced_string = re.sub(
                            contraction,
                            custom_replace,
                            replaced_string,
                            flags=re.IGNORECASE | re.DOTALL,
                        )
                        if sample.task in ("ner", "text-classification"):
                            transformations.append(
                                Transformation(
                                    original_span=Span(
                                        start=search.start(),
                                        end=search.end(),
                                        word=search.group(),
                                    ),
                                    new_span=Span(
                                        start=search.start(),
                                        end=search.end() + diff_len,
                                        word=new_string,
                                    ),
                                    ignore=False,
                                )
                            )
                sample.test_case = replaced_string
                if sample.task in ("ner", "text-classification"):
                    sample.transformations = transformations
                sample.category = "robustness"
        return sample_list


class DyslexiaWordSwap(BaseRobustness):
    """A class for simulating dyslexic word swapping."""

    alias_name = "dyslexia_word_swap"

    @staticmethod
    def transform(sample_list: List[Sample], prob: Optional[float] = 1.0) -> List[Sample]:
        """Converts the input sentences by changing some similar words from the dyslexia map and outputs a new string.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of adding dyslexia words to each sample.
                                    Defaults to 1.0, which means all samples will be transformed.

        Returns:
            List[Sample]: The transformed sample list with dyslexia words added.
        """

        def dyslexia_word_swap(text):
            transformations = []

            def replace_word(match):
                original_word = match.group()
                transformed_word = dyslexia_map.get(original_word, original_word)
                if transformed_word != original_word and (random.random() < prob):
                    transformations.append(
                        Transformation(
                            original_span=Span(
                                start=match.start(), end=match.end(), word=original_word
                            ),
                            new_span=Span(
                                start=match.start(),
                                end=match.start() + len(transformed_word),
                                word=transformed_word,
                            ),
                            ignore=False,
                        )
                    )
                    return transformed_word
                return original_word

            pattern = r"\b\w+\b"  # Matches whole words using word boundaries
            transformed_text = re.sub(pattern, replace_word, text)

            return transformed_text, transformations

        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                sample_list[idx], _ = dyslexia_word_swap(sample)
            else:
                sample.test_case, transformations = dyslexia_word_swap(sample.original)
                if sample.task in ("ner", "text-classification"):
                    sample.transformations = transformations
                sample.category = "robustness"

        return sample_list


class NumberToWord(BaseRobustness):
    """A class for converting numbers to words."""

    alias_name = "number_to_word"
    num = ConvertNumberToWord()

    @staticmethod
    def transform(sample_list: List[Sample], prob: Optional[float] = 1.0) -> List[Sample]:
        """Converts numbers in the input text to their word representations.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of converting numbers to words in each sample.
                                    Defaults to 1.0, which means all samples will be transformed.

        Returns:
            List[Sample]: The transformed sample list.
        """

        def convert_numbers(regex: str, text: str) -> Tuple[str, List[Transformation]]:
            perturbed_text = text
            transformations = []

            digit_count = len(re.findall(regex, perturbed_text))

            for _ in range(digit_count):
                matches = re.finditer(regex, perturbed_text)
                for match in matches:
                    start, end = match.start(), match.end()
                    token = perturbed_text[start:end]
                    word_num = NumberToWord.num.number_to_words(token)
                    if word_num and (random.random() < prob):
                        perturbed_text = (
                            f"{perturbed_text[:start]}{word_num}{perturbed_text[end:]}"
                        )
                        transformations.append(
                            Transformation(
                                original_span=Span(start=start, end=end, word=token),
                                new_span=Span(
                                    start=start, end=start + len(word_num), word=word_num
                                ),
                                ignore=False,
                            )
                        )
                        break

            return perturbed_text, transformations

        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                sample_list[idx], _ = convert_numbers(
                    r"(?<!\S)(\d+(\.\d+)?)(?=(\s|\n|$))", sample
                )
            else:
                sample.test_case, transformations = convert_numbers(
                    r"(?<!\S)(\d+(\.\d+)?)(?=(\s|\n|$))", sample.original
                )
                if sample.task in ("ner", "text-classification"):
                    sample.transformations = transformations
                sample.category = "robustness"
        return sample_list


class AddOcrTypo(BaseRobustness):
    """A class for adding OCR typos to the input text."""

    alias_name = "add_ocr_typo"

    @staticmethod
    def transform(
        sample_list: List[Sample], prob: Optional[float] = 1.0, count: int = 1
    ) -> List[Sample]:
        """Add OCR typos to the input samples.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of adding OCR typos to each sample.
                                    Defaults to 1.0, which means all samples will be transformed.
            count: Number of variations to create.

        Returns:
            List[Sample]: The transformed sample list with ocr typos added
        """

        def ocr_typo(regex, text):
            perturbed_text = text
            transformations = []

            for word, typo_word in inverted_ocr_typo_dict.items():
                typo_word = random.choice(typo_word)
                matches = re.finditer(regex, perturbed_text)
                for match in matches:
                    start = match.start()
                    end = match.end()
                    token = perturbed_text[start:end]
                    if token.lower() == word and (random.random() < prob):
                        if token.isupper():
                            typo_word = typo_word.upper()
                        perturbed_text = (
                            perturbed_text[:start] + typo_word + perturbed_text[end:]
                        )
                        transformations.append(
                            Transformation(
                                original_span=Span(start=start, end=end, word=token),
                                new_span=Span(
                                    start=start,
                                    end=start + len(typo_word),
                                    word=typo_word,
                                ),
                                ignore=False,
                            )
                        )

            return perturbed_text, transformations

        perturbed_samples = []
        for s in sample_list:
            for i in range(count):
                sample = deepcopy(s)
                if isinstance(sample, str):
                    sample, _ = ocr_typo(r"[^,\s.!?]+", sample)
                else:
                    sample.test_case, transformations = ocr_typo(
                        r"[^,\s.!?]+", sample.original
                    )
                    if sample.task in ("ner", "text-classification"):
                        sample.transformations = transformations
                    sample.category = "robustness"
                perturbed_samples.append(sample)
        return perturbed_samples


class AbbreviationInsertion(BaseRobustness):
    """A class for adding abbreviations to the input text."""

    alias_name = "add_abbreviation"

    @staticmethod
    def transform(sample_list: List[Sample], prob: Optional[float] = 1.0) -> List[Sample]:
        """Transforms the given sample list by inserting abbreviations.

        Args:
            sample_list (List[Sample]): The list of samples to transform.
            prob (Optional[float]): The probability controlling the proportion of words to be perturbed.
                                    Defaults to 1.0, which means all samples will be transformed.

        Returns:
            List[Sample]: The transformed list of samples with abbreviations added
        """

        def insert_abbreviation(text):
            perturbed_text = text
            transformations = []

            for abbreviation, expansions in abbreviation_dict.items():
                for expansion in expansions:
                    pattern = r"(?i)\b" + re.escape(expansion) + r"\b"
                    corrected_token = abbreviation
                    matches = re.finditer(pattern, perturbed_text)
                    for match in matches:
                        start = match.start()
                        end = match.end()
                        token = perturbed_text[start:end]
                        if corrected_token != token and (random.random() < prob):
                            perturbed_text = (
                                perturbed_text[:start]
                                + corrected_token
                                + perturbed_text[end:]
                            )
                            transformations.append(
                                Transformation(
                                    original_span=Span(start=start, end=end, word=token),
                                    new_span=Span(
                                        start=start,
                                        end=start + len(corrected_token),
                                        word=corrected_token,
                                    ),
                                    ignore=False,
                                )
                            )

            return perturbed_text, transformations

        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                sample_list[idx], _ = insert_abbreviation(sample)
            else:
                sample.test_case, transformations = insert_abbreviation(sample.original)
                if sample.task in ("ner", "text-classification"):
                    sample.transformations = transformations
                sample.category = "robustness"

        return sample_list


class AddSpeechToTextTypo(BaseRobustness):
    """A class for adding common speech to text typos to the input text."""

    alias_name = "add_speech_to_text_typo"

    @staticmethod
    def transform(
        sample_list: List[Sample], prob: Optional[float] = 1.0, count: int = 1
    ) -> List[Sample]:
        """Transforms the given sample list by introducing typos simulating speech-to-text errors.

        Args:
            sample_list (List[Sample]): The list of samples to transform.
            prob (Optional[float]): The probability controlling the proportion of words to be perturbed.
                                    Defaults to 1.0, which means all samples will be transformed.
            count: Number of variations to create.

        Returns:
            List[Sample]: The transformed list of samples with speech to text typos added.
        """

        def convertToSimilarHarmony(sentence):
            perturbed_text = sentence
            transformations = []

            for match in re.finditer(r"\w+(?:'\w+)*|\W", perturbed_text):
                start, end = match.start(), match.end()
                word = perturbed_text[start:end]
                try:
                    if word.isalpha() and (random.random() < prob):
                        # get similar words from the dictionary
                        similar_words = Search.perfectHomophones(word)

                        if similar_words:
                            new_word = random.choice(similar_words)

                            # check if the word is the same as the original word
                            if word.lower() == new_word.lower():
                                continue

                            # check if the word or first letter in sentence is capitalized or not
                            if word[0].isupper() or not start:
                                new_word = new_word.capitalize()

                            perturbed_text = (
                                perturbed_text[:start] + new_word + perturbed_text[end:]
                            )

                            transformations.append(
                                Transformation(
                                    original_span=Span(start=start, end=end, word=word),
                                    new_span=Span(
                                        start=start,
                                        end=start + len(new_word),
                                        word=new_word,
                                    ),
                                    ignore=False,
                                )
                            )
                            break

                except ValueError:
                    # if the word is not in the dictionary, skip it
                    continue

            return perturbed_text, transformations

        perturbed_samples = []
        for s in sample_list:
            for i in range(count):
                sample = deepcopy(s)
                if isinstance(sample, str):
                    sample, _ = convertToSimilarHarmony(sample)
                else:
                    sample.test_case, transformations = convertToSimilarHarmony(
                        sample.original
                    )
                    if sample.task in ("ner", "text-classification"):
                        sample.transformations = transformations
                    sample.category = "robustness"
                perturbed_samples.append(sample)
        return perturbed_samples


class AddSlangifyTypo(BaseRobustness):
    """A class for adding slangs to text typos to the input text."""

    alias_name = "add_slangs"

    @staticmethod
    def transform(sample_list: List[Sample], prob: Optional[float] = 1.0) -> List[Sample]:
        """Transforms the given sample list by adding slang words.

        Args:
            sample_list (List[Sample]): The list of samples to transform.
            prob (Optional[float]): The probability controlling the proportion of words to be perturbed.
                                    Defaults to 1.0, which means all samples will be transformed.

        Returns:
            List[Sample]: The transformed list of samples with slangs added.
        """

        def slangify_typo(text):
            slang_words = [
                list(map(list, zip(*Slang_Nouns))),
                list(map(list, zip(*Slang_Adverbs))),
                list(map(list, zip(*Slang_Adjectives))),
            ]

            modified_toks = []
            tokens = re.findall(
                r"\w+(?:[-']\w+)*|[^\w\s]|[\s]+", text
            )  # Include hyphenated and possessive words as single tokens
            transformations = []
            start_offset = 0

            for token in tokens:
                if token.isspace() or all(ch in string.punctuation for ch in token):
                    modified_toks.append(
                        token
                    )  # Preserve space and punctuation in the reconstructed text
                    continue
                is_cap = token[0].isupper()
                replaced = False

                for slang in slang_words:
                    if token.lower() in slang[0]:
                        replacements = [
                            i for i, x in enumerate(slang[0]) if x == token.lower()
                        ]
                        chosen_index = random.choice(replacements)
                        temp = slang[1][chosen_index]

                        if is_cap:
                            temp = temp[0].upper() + temp[1:]

                        if temp != token and (random.random() < prob):
                            start_index = text.index(token, start_offset)
                            end_index = start_index + len(token)
                            modified_toks.append(temp)
                            new_word_length = len(temp)
                            transformations.append(
                                Transformation(
                                    original_span=Span(
                                        start=start_index, end=end_index, word=token
                                    ),
                                    new_span=Span(
                                        start=start_index,
                                        end=start_index + new_word_length,
                                        word=temp,
                                    ),
                                    ignore=False,
                                )
                            )
                            start_offset = end_index
                        else:
                            modified_toks.append(token)

                        replaced = True
                        break

                if not replaced:
                    modified_toks.append(token)

            modified_text = "".join(modified_toks)

            return modified_text, transformations

        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                sample_list[idx], _ = slangify_typo(sample)
            else:
                sample.test_case, transformations = slangify_typo(sample.original)
                if sample.task in ("ner", "text-classification"):
                    sample.transformations = transformations
                sample.category = "robustness"

        return sample_list


class MultiplePerturbations(BaseRobustness):
    """A class for applying multiple perturbations to transform a sample list."""

    alias_name = "multiple_perturbations"

    @staticmethod
    def transform(
        sample_list: List[Sample],
        perturbations: List[str],
        prob: Optional[float] = 1.0,
        config=None,
    ) -> List[Sample]:
        """Transforms the given sample list by applying multiple perturbations.

        Args:
            sample_list (List[Sample]): The list of samples to transform.
            perturbations (List[str]): The list of perturbations to apply.
            prob (Optional[float]): The probability controlling the proportion of words to be perturbed.
                                    Defaults to 1.0, which means all samples will be transformed.

        Returns:
            transformed_list: The transformed list of samples.
        """

        def apply_transformation(sample, order, prob):
            if order == "uppercase":
                transformed_list = UpperCase.transform(sample, prob)
            elif order == "lowercase":
                transformed_list = LowerCase.transform(sample, prob)
            elif order == "titlecase":
                transformed_list = TitleCase.transform(sample, prob)
            elif order == "add_punctuation":
                transformed_list = AddPunctuation.transform(sample, prob)
            elif order == "strip_punctuation":
                transformed_list = StripPunctuation.transform(sample, prob)
            elif order == "add_typo":
                transformed_list = AddTypo.transform(sample, prob)
            elif order == "american_to_british":
                transformed_list = ConvertAccent.transform(
                    sample,
                    prob,
                    **config.get("american_to_british", {}).get("parameters", {}),
                )
            elif order == "british_to_american":
                transformed_list = ConvertAccent.transform(
                    sample,
                    prob,
                    **config.get("british_to_american", {}).get("parameters", {}),
                )
            elif next(iter(order)) == "add_context":
                transformed_list = AddContext.transform(
                    sample,
                    prob,
                    order["add_context"]["parameters"]["starting_context"],
                    order["add_context"]["parameters"]["ending_context"],
                )
            elif order == "add_contraction":
                transformed_list = AddContraction.transform(sample, prob)
            elif order == "dyslexia_word_swap":
                transformed_list = DyslexiaWordSwap.transform(sample, prob)
            elif order == "number_to_word":
                transformed_list = NumberToWord.transform(sample, prob)
            elif order == "add_abbreviation":
                transformed_list = AbbreviationInsertion.transform(sample, prob)
            elif order == "add_ocr_typo":
                transformed_list = AddOcrTypo.transform(sample, prob)
            elif order == "add_speech_to_text_typo":
                transformed_list = AddSpeechToTextTypo.transform(sample, prob)
            elif order == "add_slangs":
                transformed_list = AddSlangifyTypo.transform(sample, prob)
            elif order == "adjective_synonym_swap":
                transformed_list = AdjectiveSynonymSwap.transform(sample, prob)
            elif order == "adjective_antonym_swap":
                transformed_list = AdjectiveAntonymSwap.transform(sample, prob)
            elif order == "strip_all_punctuation":
                transformed_list = StripAllPunctuation.transform(sample, prob)
            else:
                raise ValueError(f"Unknown transformation: {order}")
            return transformed_list

        if isinstance(sample_list[0], SequenceClassificationSample):
            for idx, transformation in enumerate(perturbations):
                if idx == 0:
                    transformed_list = apply_transformation(
                        sample_list, transformation, prob
                    )
                else:
                    new_list = []
                    for sample in transformed_list:
                        new_sample = SequenceClassificationSample(
                            original=sample.test_case,
                            category="robustness",
                            expected_results=sample.expected_results,
                        )
                        new_list.append(new_sample)
                    transformed_list = apply_transformation(
                        new_list, perturbations[idx], prob
                    )

            for i, sample in enumerate(transformed_list):
                sample.original = sample_list[i].original
                sample.transformations = None

        elif isinstance(sample_list[0], str):
            for idx, transformation in enumerate(perturbations):
                transformed_list = apply_transformation(sample_list, transformation, prob)

        return transformed_list


class AdjectiveSynonymSwap(BaseRobustness):
    """A class for swaping adjectives to their synonyms"""

    alias_name = "adjective_synonym_swap"

    @staticmethod
    def transform(sample_list: List[Sample], prob: Optional[float] = 1.0) -> List[Sample]:
        """Converts the input sentences by changing adjectives to their respective synonyms from the adjective synonym dictionary.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of replacing adjectives words to their respective synonym for each sample.
                                    Defaults to 1.0, which means all samples will be transformed.

        Returns:
            List[Sample]: The transformed sample list with adjectives swapped with their respective synonyms.
        """

        def adjective_synonym_swap(text):
            transformations = []

            def replace_word(match):
                original_word = match.group()
                synonyms = adjective_synonym_dict.get(original_word, [original_word])
                transformed_word = random.choice(synonyms)
                if transformed_word != original_word and (random.random() < prob):
                    transformations.append(
                        Transformation(
                            original_span=Span(
                                start=match.start(),
                                end=match.end(),
                                word=original_word,
                            ),
                            new_span=Span(
                                start=match.start(),
                                end=match.start() + len(transformed_word),
                                word=transformed_word,
                            ),
                            ignore=False,
                        )
                    )
                    return transformed_word
                return original_word

            pattern = r"\b\w+\b"  # Matches whole words using word boundaries
            transformed_text = re.sub(pattern, replace_word, text)

            return transformed_text, transformations

        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                sample_list[idx], _ = adjective_synonym_swap(sample)
            else:
                sample.test_case, transformations = adjective_synonym_swap(
                    sample.original
                )
                if sample.task in ("ner", "text-classification"):
                    sample.transformations = transformations
                sample.category = "robustness"

        return sample_list


class AdjectiveAntonymSwap(BaseRobustness):
    """A class for swaping adjectives to their antonyms"""

    alias_name = "adjective_antonym_swap"

    @staticmethod
    def transform(sample_list: List[Sample], prob: Optional[float] = 1.0) -> List[Sample]:
        """Converts the input sentences by changing adjectives to their respective antonyms from the adjective antonym dictionary.

        Args:
            sample_list (List[Sample]): A list of samples to be transformed.
            prob (Optional[float]): The probability of replacing adjectives words to their respective antonyms for each sample.
                                    Defaults to 1.0, which means all samples will be transformed.

        Returns:
            List[Sample]: The transformed sample list with adjectives swapped with their respective antonyms.
        """

        def adjective_antonym_swap(text):
            transformations = []

            def replace_word(match):
                original_word = match.group()
                antonyms = adjective_antonym_dict.get(original_word, [original_word])
                transformed_word = random.choice(antonyms)
                if transformed_word != original_word and (random.random() < prob):
                    transformations.append(
                        Transformation(
                            original_span=Span(
                                start=match.start(),
                                end=match.end(),
                                word=original_word,
                            ),
                            new_span=Span(
                                start=match.start(),
                                end=match.start() + len(transformed_word),
                                word=transformed_word,
                            ),
                            ignore=False,
                        )
                    )
                    return transformed_word
                return original_word

            pattern = r"\b\w+\b"  # Matches whole words using word boundaries
            transformed_text = re.sub(pattern, replace_word, text)

            return transformed_text, transformations

        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                sample_list[idx], _ = adjective_antonym_swap(sample)
            else:
                sample.test_case, transformations = adjective_antonym_swap(
                    sample.original
                )
                if sample.task in ("ner", "text-classification"):
                    sample.transformations = transformations
                sample.category = "robustness"

        return sample_list


class StripAllPunctuation(BaseRobustness):
    """A class for stripping punctuation from text samples."""

    alias_name = "strip_all_punctuation"

    @staticmethod
    def transform(
        sample_list: List[Union[Sample, str]],
        prob: Optional[float] = 1.0,
        whitelist: Optional[List[str]] = None,
    ) -> List[Sample]:
        """
        Transforms the given sample list by stripping all punctuations.

        Args:
            sample_list (List[Union[Sample, str]]): The list of samples to transform.
            prob (Optional[float]): The probability controlling the proportion of samples to be perturbed.
                                    Defaults to 1.0, which means all samples will be transformed.
            whitelist (Optional[List[str]]): Punctuations to look for.

        Returns:
            transformed_list: The transformed list of samples.
        """
        if whitelist is None:
            whitelist = ["!", "?", ",", ".", "-", ":", ";", "/", "'", '"']

        exceptions = ["s/p", "h/o"]
        letter_letter_pattern = r"\b\w/\w\b"
        number_pattern = r"\b\d+\.\d+\b|\b\d{1,2}/\d{1,2}/\d{2,4}\b|\b\d{1,2}-\d{1,2}-\d{2,4}\b|\b\d+-\d+\b"

        exceptions_pattern = "|".join(
            [
                f"(?<!{ex.split('/')[0]})/(?!{ex.split('/')[1]})"
                for ex in exceptions
                if "/" in ex
            ]
        )
        whitelist_pattern = "|".join(
            [f"\\{char}" for char in whitelist if char not in ["/", ".", "-"]]
        )

        pattern = "|".join(
            [
                number_pattern,  # to handle decimal numbers
                exceptions_pattern,
                whitelist_pattern,
                letter_letter_pattern,  # to handle letter/letter
                "(?<!\\d)\\.(?!\\d)",  # to handle non-decimal periods
                "(?<!\\d)-(?!\\d)",  # to handle non-range hyphens
            ]
        )

        def check_whitelist(text):
            new_text = text
            transformations = []
            offset = 0
            for match in re.finditer(pattern, new_text):
                if (
                    re.match(letter_letter_pattern, match.group())
                    and match.group() not in exceptions
                ):
                    transformations.append(
                        Transformation(
                            original_span=Span(
                                start=match.start() - offset,
                                end=match.end() - offset,
                                word=match.group(),
                            ),
                            new_span=Span(
                                start=match.start() - offset,
                                end=match.start() - offset + 5,
                                word=" and ",
                            ),
                        )
                    )
                    new_text = (
                        new_text[: match.start() - offset]
                        + " and "
                        + new_text[match.end() - offset :]
                    )
                    offset += 1
                elif not re.match(
                    number_pattern, match.group()
                ):  # Avoid removing punctuation from decimal numbers, date formats, and ranges
                    transformations.append(
                        Transformation(
                            original_span=Span(
                                start=match.start() - offset,
                                end=match.end() - offset,
                                word=match.group(),
                            ),
                            new_span=Span(
                                start=match.start() - offset,
                                end=match.start() - offset,
                                word="",
                            ),
                        )
                    )
                    new_text = (
                        new_text[: match.start() - offset]
                        + new_text[match.end() - offset :]
                    )
                    offset += len(match.group())

            return new_text, transformations

        for idx, sample in enumerate(sample_list):
            if isinstance(sample, str):
                if random.random() < prob:
                    transformed_text, transformations = check_whitelist(sample)
                    sample_list[idx] = transformed_text
            else:
                if random.random() < prob:
                    transformed_text, transformations = check_whitelist(sample.original)
                    sample.test_case = transformed_text
                    if sample.task in ("ner", "text-classification"):
                        sample.transformations = transformations
                else:
                    sample.test_case = sample.original

                sample.category = "robustness"

        return sample_list


class RandomAge(BaseRobustness):
    """A class for adding abbreviations to the input text."""

    alias_name = "randomize_age"

    @staticmethod
    def transform(
        sample_list: List[Sample],
        prob: Optional[float] = 1.0,
        random_amount: int = 5,
        count: int = 1,
    ) -> List[Sample]:
        """Transforms the given sample list by randomizing the ages by a certain amount.

        Args:
            sample_list (List[Sample]): The list of samples to transform.
            prob (Optional[float]): The probability controlling the proportion of words to be perturbed.
                                    Defaults to 1.0, which means all samples will be transformed.
            random_amount (Optional[int]): The range to randomize the ages. +-random_amount is added to ages.
                                    Default is 5.
            count (Optional[int]): Number of variations to create from one sample.

        Returns:
            List[Sample]: The transformed list of samples with abbreviations added
        """
        age_expressions = [
            r"\d+ years old",
            r"\d+ months old",
            r"\d+ weeks old",
            r"\d+ days old",
        ]

        def randomize_ages(text):
            perturbed_text = text
            transformations = []

            for expr in age_expressions:
                matches = re.finditer(expr, text)
                for match in matches:
                    start = match.start()
                    end = match.end()
                    token = text[start:end]
                    new_age = random.randint(-random_amount, random_amount) + int(
                        token.split(" ")[0]
                    )
                    new_age = new_age if new_age > 0 else 1
                    corrected_token = re.sub(r"\d+", str(new_age), token)
                    if corrected_token != token and (random.random() < prob):
                        perturbed_text = (
                            perturbed_text[:start]
                            + corrected_token
                            + perturbed_text[end:]
                        )
                        transformations.append(
                            Transformation(
                                original_span=Span(start=start, end=end, word=token),
                                new_span=Span(
                                    start=start,
                                    end=start + len(corrected_token),
                                    word=corrected_token,
                                ),
                                ignore=False,
                            )
                        )

            return perturbed_text, transformations

        perturbed_samples = []
        for sample in sample_list:
            for i in range(count):
                if isinstance(sample, str):
                    s, _ = randomize_ages(sample)
                    perturbed_samples.append(s)
                else:
                    s = deepcopy(sample)
                    s.test_case, transformations = randomize_ages(s.original)
                    if s.task in ("ner", "text-classification"):
                        s.transformations = transformations
                    s.category = "robustness"

        return perturbed_samples



from typing import Tuple, List, Union
import torch
import numpy as np
from transformers import AutoModel, AutoTokenizer


class SimpleSentenceTransformer:
    """A simple class to handle the sentence transformation using the specified model.

    Attributes:
        device (torch.device): The device used for computations, i.e., either a GPU (if available) or a CPU.
        tokenizer (transformers.AutoTokenizer): The tokenizer associated with the model.
        model (transformers.AutoModel): The transformer model used for sentence embeddings.
    """

    def __init__(
        self,
        model_name: str,
    ):
        """Constructor method

        Args:
            model_name (str): The name of the model to be loaded. By default, it uses the multilingual MiniLM model.
        """
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModel.from_pretrained(model_name).to(self.device)

    def mean_pooling(
        self, model_output: Tuple[torch.Tensor], attention_mask: torch.Tensor
    ) -> torch.Tensor:
        """Apply mean pooling on the model outputs.

        Args:
            model_output (torch.Tensor): The model's output.
            attention_mask (torch.Tensor): The attention mask tensor.

        Returns:
            torch.Tensor: The mean pooled output tensor.
        """
        token_embeddings = model_output[
            0
        ]  # First element of model_output contains all token embeddings
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
            input_mask_expanded.sum(1), min=1e-9
        )

    def encode(
        self,
        sentences: Union[str, List[str]],
        convert_to_tensor: bool = False,
        max_length: int = 128,
    ) -> Union[torch.Tensor, np.ndarray]:
        """Encode sentences into sentence embeddings.

        Args:
            sentences (str | list): The sentences to be encoded. Can be either a single string or a list of strings.
            convert_to_tensor (bool, optional): If set to True, the method will return tensors, otherwise it will return numpy arrays. Defaults to False.
            max_length (int, optional): The maximum length for the sentences. Any sentence exceeding this length gets truncated. Defaults to 128.

        Returns:
            torch.Tensor | numpy.ndarray: The sentence embeddings. The datatype depends on the 'convert_to_tensor' parameter.
        """
        # Ensure sentences is a list
        if not isinstance(sentences, list):
            sentences = [sentences]

        # Tokenize the sentences
        encoded_input = self.tokenizer(
            sentences,
            padding="max_length",
            truncation=True,
            max_length=max_length,
            return_tensors="pt",
        ).to(self.device)

        # Get the model's output
        with torch.no_grad():
            model_output = self.model(**encoded_input)

        # Perform pooling. In this case, mean pooling.
        sentence_embeddings = self.mean_pooling(
            model_output, encoded_input["attention_mask"]
        )

        if convert_to_tensor:
            return sentence_embeddings
        else:
            return sentence_embeddings.cpu().numpy()



from typing import Optional

from pydantic import BaseModel, Field

from .helpers import Span


class NERPrediction(BaseModel):
    """Single prediction obtained from a named entity recognition model"""

    entity: str = Field(None, alias="entity_group")
    span: Span
    score: Optional[float] = None
    doc_id: Optional[int] = None
    doc_name: Optional[str] = None
    pos_tag: Optional[str] = None
    chunk_tag: Optional[str] = None

    class Config:
        extra = "ignore"
        allow_population_by_field_name = True

    @classmethod
    def from_span(
        cls,
        entity: str,
        word: str,
        start: int,
        end: int,
        score: float = None,
        doc_id: int = None,
        doc_name: str = None,
        pos_tag: str = None,
        chunk_tag: str = None,
    ) -> "NERPrediction":
        """"""
        return cls(
            entity=entity,
            span=Span(start=start, end=end, word=word),
            score=score,
            doc_id=doc_id,
            doc_name=doc_name,
            pos_tag=pos_tag,
            chunk_tag=chunk_tag,
        )

    def __hash__(self):
        """"""
        return hash(self.__repr__())

    def __eq__(self, other):
        """"""
        if isinstance(other, NERPrediction):
            return (
                self.entity == other.entity
                and self.span.start == other.span.start
                and self.span.end == other.span.end
            )
        return False

    def __str__(self) -> str:
        """"""
        return f"{self.span.word}: {self.entity}"

    def __repr__(self) -> str:
        """"""
        return f"<NERPrediction(entity='{self.entity}', span={self.span})>"


class SequenceLabel(BaseModel):
    """Single prediction obtained from text-classification models"""

    label: str
    score: float

    def __str__(self):
        """"""
        return f"{self.label}"



from abc import ABC, abstractmethod
import asyncio
from typing import List
from langtest.modelhandler.modelhandler import ModelFactory

from langtest.utils.custom_types.sample import Sample


class BaseSecurity(ABC):
    """Abstract base class for implementing a model security.

    This class defines the interface for implementing a model security.

    Attributes:
        None
    """

    @staticmethod
    @abstractmethod
    def transform():
        """Abstract method that transforms the sample data based on the implemented model security."""
        raise NotImplementedError("Please Implement this method")

    @staticmethod
    @abstractmethod
    async def run(sample_list: List[Sample], model: ModelFactory, **kwargs):
        """Abstract method that implements the model security."""
        progress = kwargs.get("progress_bar", False)
        for sample in sample_list:
            if sample.state != "done":
                if hasattr(sample, "run"):
                    sample_status = sample.run(model, **kwargs)
                    if sample_status:
                        sample.state = "done"
                else:
                    sample.actual_results = model(sample.prompt)
                    sample.state = "done"
            if progress:
                progress.update(1)
        return sample_list

    @classmethod
    async def async_run(cls, sample_list: List[Sample], model: ModelFactory, **kwargs):
        """Abstract method that implements the model security."""
        created_task = await asyncio.create_task(cls.run(sample_list, model, **kwargs))
        return created_task


class PromptInjection(BaseSecurity):
    """
    PromptInjection is a class that implements the model security for prompt injection.
    """

    alias_name = ["prompt_injection_attack"]
    supported_tasks = ["security"]

    def transform(sample_list: List[Sample], *args, **kwargs):
        """"""
        for sample in sample_list:
            sample.test_type = "prompt_injection_attack"
            sample.category = "security"

        return sample_list



from typing import List, Optional, TypeVar, Union
from pydantic import BaseModel, validator
from .helpers import Span
from .predictions import NERPrediction, SequenceLabel


class SequenceClassificationOutput(BaseModel):
    """Output model for text classification tasks."""

    predictions: List[SequenceLabel]

    def to_str_list(self) -> str:
        """Convert the output into list of strings.

        Returns:
            List[str]: predictions in form of a list of strings.
        """
        return ",".join([x.label for x in self.predictions])

    def __str__(self):
        """String representation"""
        labels = {elt.label: elt.score for elt in self.predictions}
        return f"SequenceClassificationOutput(predictions={labels})"

    def __eq__(self, other):
        """Equality comparison method."""
        top_class = max(self.predictions, key=lambda x: x.score).label
        other_top_class = max(other.predictions, key=lambda x: x.score).label
        return top_class == other_top_class


class MinScoreOutput(BaseModel):
    """Output for accuracy/representation tests."""

    min_score: float

    def to_str_list(self) -> float:
        """Convert the output into list of strings.

        Returns:
            List[str]: predictions in form of a list of strings.
        """
        return self.min_score

    def __repr__(self) -> str:
        """Printable representation"""
        return f"{self.min_score}"

    def __str__(self) -> str:
        """String representation"""
        return f"{self.min_score}"


class MaxScoreOutput(BaseModel):
    """Output for accuracy/representation tests."""

    max_score: float

    def to_str_list(self) -> float:
        """Formatting helper"""
        return self.max_score

    def __repr__(self) -> str:
        """Printable representation"""
        return f"{self.max_score}"

    def __str__(self) -> str:
        """String representation"""
        return f"{self.max_score}"


class NEROutput(BaseModel):
    """Output model for NER tasks."""

    predictions: List[NERPrediction]

    @validator("predictions")
    def sort_by_appearance(cls, v):
        """Sort spans by order of appearance in the text"""
        return sorted(v, key=lambda x: x.span.start)

    def __len__(self):
        """Number of detected entities"""
        return len(self.predictions)

    def __getitem__(
        self, item: Union[Span, int, str]
    ) -> Optional[Union[List[NERPrediction], NERPrediction]]:
        """Item getter"""
        if isinstance(item, int):
            return self.predictions[item]
        elif isinstance(item, str):
            for pred in self.predictions:
                if pred.span.word == item:
                    return pred
            return None
        elif isinstance(item, Span):
            for prediction in self.predictions:
                if prediction.span == item:
                    return prediction
            return None
        elif isinstance(item, slice):
            return [self.predictions[i] for i in range(item.start, item.stop)]

    def to_str_list(self) -> str:
        """Converts predictions into a list of strings.

        Returns:
            List[str]: predictions in form of a list of strings.
        """
        return ", ".join([str(x) for x in self.predictions if str(x)[-3:] != ": O"])

    def __repr__(self) -> str:
        """Printable representation"""
        return self.predictions.__repr__()

    def __str__(self) -> str:
        """String representation"""
        return [str(x) for x in self.predictions].__repr__()

    def __eq__(self, other: "NEROutput"):
        """Equality comparison method."""
        # NOTE: we need the list of transformations applied to the sample to be able
        # to align and compare two different NEROutput
        raise NotImplementedError()


class TranslationOutput(BaseModel):
    """Output model for translation tasks."""

    translation_text: str  # Changed from List[str] to str

    def to_str_list(self) -> List[str]:
        """Formatting helper

        Returns:
             List[str]: the translation_text as a list of strings.
        """
        return [self.translation_text]  # Wrap self.translation_text in a list

    def __str__(self):
        """String representation of TranslationOutput."""
        return self.translation_text  # Return translation_text directly

    def __eq__(self, other):
        """Equality comparison method."""
        if isinstance(other, TranslationOutput):
            return self.translation_text == other.translation_text
        if isinstance(other, list):
            return [self.translation_text] == other
        return False


Result = TypeVar(
    "Result",
    NEROutput,
    SequenceClassificationOutput,
    MinScoreOutput,
    TranslationOutput,
    MaxScoreOutput,
    List[str],
    str,
)



from .constants import (
    asian_names,
    black_names,
    country_economic_dict,
    hispanic_names,
    inter_racial_names,
    male_pronouns,
    native_american_names,
    neutral_pronouns,
    religion_wise_names,
    white_names,
    female_pronouns,
)
from typing import Union


def add_custom_data(data: Union[list, dict], name: str, append: bool) -> None:
    """
    Adds custom data to the corresponding bias dictionaries based on the specified name.

    Args:
        data (dict or list): The data to be added.
        name (str): The type of dictionary to update. It should be one of the following:
            - "Country-Economic-Bias"
            - "Religion-Bias"
            - "Ethnicity-Name-Bias"
            - "Gender-Pronoun-Bias"
            - "Country-Economic-Representation"
            - "Religion-Representation"
            - "Ethnicity-Representation"
        append (bool): Specifies whether to append the values or overwrite them.

    Raises:
        ValueError: If the specified `name` is invalid or if the provided data has an invalid format or contains invalid keys.
    """
    if name in ("Country-Economic-Bias", "Country-Economic-Representation"):
        valid_names = country_economic_dict.keys()

        # Validate the schema
        if not set(data.keys()).issubset(valid_names):
            raise ValueError(
                f"Invalid schema. It should be one of: {', '.join(valid_names)}."
            )

        if append:
            # Append unique values to existing keys
            for key, values in data.items():
                unique_values = set(values)
                country_economic_dict[key] = list(
                    set(country_economic_dict[key]) | unique_values
                )
        else:
            # Overwrite the keys' values
            for key, values in data.items():
                country_economic_dict[key] = values

    elif name in ("Religion-Bias", "Religion-Representation"):
        valid_names = religion_wise_names.keys()

        # Validate the schema
        if not set(data.keys()).issubset(valid_names):
            raise ValueError(
                f"Invalid schema. It should be one of: {', '.join(valid_names)}."
            )

        if append:
            # Append unique values to existing keys
            for key, values in data.items():
                unique_values = set(values)
                religion_wise_names[key] = list(
                    set(religion_wise_names[key]) | unique_values
                )
        else:
            # Overwrite the keys' values
            for key, values in data.items():
                religion_wise_names[key] = values

    elif name in ("Ethnicity-Name-Bias", "Ethnicity-Representation"):
        ethnicity_data = {
            "white_names": white_names,
            "black_names": black_names,
            "hispanic_names": hispanic_names,
            "asian_names": asian_names,
            "native_american_names": native_american_names,
            "inter_racial_names": inter_racial_names,
        }

        valid_names = tuple(ethnicity_data.keys())

        for data_dict in data:
            if "name" not in data_dict:
                raise ValueError("Invalid JSON format. 'name' key is missing.")

            name = data_dict["name"]
            first_names = data_dict.get("first_names", [])
            last_names = data_dict.get("last_names", [])

            if not isinstance(name, str):
                raise ValueError("Invalid 'name' format in the JSON file.")

            if name not in valid_names:
                raise ValueError(
                    f"Invalid 'name' value '{name}'. It should be one of: {', '.join(valid_names)}."
                )

            if not first_names and not last_names:
                if name not in ("native_american_names", "inter_racial_names"):
                    raise ValueError(
                        f"At least one of 'first_names' or 'last_names' must be specified for '{name}'."
                    )
                else:
                    raise ValueError(f"'last_names' must be specified for '{name}'.")

            if set(data_dict.keys()) - {"name", "first_names", "last_names"}:
                raise ValueError(
                    f"Invalid keys in the JSON for '{name}'. "
                    f"Only the following keys are allowed: 'name', 'first_names', 'last_names'."
                )

            if name in (
                "white_names",
                "black_names",
                "hispanic_names",
                "asian_names",
            ):
                if append:
                    ethnicity_data[name]["first_names"] += first_names
                    ethnicity_data[name]["last_names"] += last_names
                else:
                    ethnicity_data[name]["first_names"] = first_names
                    ethnicity_data[name]["last_names"] = last_names
            elif name in ("native_american_names", "inter_racial_names"):
                if append:
                    ethnicity_data[name]["last_names"] += last_names
                else:
                    ethnicity_data[name]["last_names"] = last_names

    elif name == "Gender-Pronoun-Bias":
        valid_names = ("female_pronouns", "male_pronouns", "neutral_pronouns")

        # Validate the schema
        for data_dict in data:
            if "name" not in data_dict:
                raise ValueError("Invalid JSON format. 'name' key is missing.")

            name = data_dict["name"]

            if name not in valid_names:
                raise ValueError(
                    f"Invalid 'name' value '{name}'. It should be one of: {', '.join(valid_names)}."
                )

            pronouns = {
                "subjective_pronouns": data_dict.get("subjective_pronouns", []),
                "objective_pronouns": data_dict.get("objective_pronouns", []),
                "reflexive_pronouns": data_dict.get("reflexive_pronouns", []),
                "possessive_pronouns": data_dict.get("possessive_pronouns", []),
            }

            if all(
                key not in pronouns
                for key in [
                    "subjective_pronouns",
                    "objective_pronouns",
                    "reflexive_pronouns",
                    "possessive_pronouns",
                ]
            ):
                raise ValueError(
                    f"Missing pronoun keys in the JSON for '{name}'. Please include at least one of: "
                    "'subjective_pronouns', 'objective_pronouns', 'reflexive_pronouns', 'possessive_pronouns'."
                )

            invalid_keys = set(data_dict.keys()) - {
                "name",
                "subjective_pronouns",
                "objective_pronouns",
                "reflexive_pronouns",
                "possessive_pronouns",
            }
            if invalid_keys:
                raise ValueError(
                    f"Invalid keys in the JSON for '{name}': {', '.join(invalid_keys)}. "
                    f"Only the following keys are allowed: "
                    "'name', 'subjective_pronouns', 'objective_pronouns', 'reflexive_pronouns', 'possessive_pronouns'."
                )

            bias_dict = {
                "female_pronouns": female_pronouns,
                "male_pronouns": male_pronouns,
                "neutral_pronouns": neutral_pronouns,
            }

            if name in bias_dict:
                bias = bias_dict[name]
                if append:
                    bias["subjective_pronouns"].extend(
                        set(pronouns["subjective_pronouns"])
                        - set(bias["subjective_pronouns"])
                    )
                    bias["objective_pronouns"].extend(
                        set(pronouns["objective_pronouns"])
                        - set(bias["objective_pronouns"])
                    )
                    bias["reflexive_pronouns"].extend(
                        set(pronouns["reflexive_pronouns"])
                        - set(bias["reflexive_pronouns"])
                    )
                    bias["possessive_pronouns"].extend(
                        set(pronouns["possessive_pronouns"])
                        - set(bias["possessive_pronouns"])
                    )
                else:
                    bias["subjective_pronouns"] = pronouns["subjective_pronouns"]
                    bias["objective_pronouns"] = pronouns["objective_pronouns"]
                    bias["reflexive_pronouns"] = pronouns["reflexive_pronouns"]
                    bias["possessive_pronouns"] = pronouns["possessive_pronouns"]



from ..transform.constants import male_names, female_names


class GenderClassifier:
    """Helper model to predict the 'gender' of a piece of text."""

    def predict(self, sentence):
        # Convert the sentence to lowercase for case-insensitive matching
        sentence = sentence.lower().strip().split()

        # Define lists of keywords for each category
        female_keywords = [
            "mrs.",
            "miss",
            "ms.",
            "she",
            "her",
            "hers",
            "woman",
            "female",
            "girl",
            "sister",
            "actress",
            "waitress",
            "stewardess",
            "princess",
            "queen",
            "duchess",
            "niece",
            "mother",
            "daughter",
            "sister",
            "madam",
            "wife",
            "bride",
            "lady",
            "female",
        ]
        female_keywords.extend(female_names)

        male_keywords = [
            "mr.",
            "he",
            "him",
            "his",
            "man",
            "male",
            "boy",
            "brother",
            "actor",
            "waiter",
            "steward",
            "prince",
            "king",
            "duke",
            "nephew",
            "father",
            "son",
            "husband",
            "groom",
            "gentleman",
            "male",
        ]
        male_keywords.extend(male_names)

        # Count the number of female and male keywords in the sentence
        female_count = sum(1 for keyword in female_keywords if keyword in sentence)
        male_count = sum(1 for keyword in male_keywords if keyword in sentence)

        # Determine the predominant gender mentioned in the sentence
        if female_count > male_count:
            return "female"
        elif male_count > female_count:
            return "male"
        else:
            return "unknown"



from .datahandler.datasource import DataFactory
from .modelhandler.modelhandler import ModelFactory
from .langtest import Harness

"""
langtest is python library package which can useful for testing of nlp models
Like Spacy, HuggingFace, SparkNLP ...etc

To testing of NLP Models by import this library as follows

Harness is a class, which can be instaniting and do testing in flow like
workflow: generate --> run --> report --> save
>>> from langtest import Harness

ModelFactory is for handling of strings to models
like access or download resources from cloud.
>>> from langtest import ModelFactory

DataFactory is for handling of like csv, json, conll ...
>>> from langtest import DataFactory


"""
__all__ = [Harness, ModelFactory, DataFactory]



import asyncio
from abc import ABC, abstractmethod
from typing import List, Dict, Union

from langtest.modelhandler.modelhandler import ModelFactory
from langtest.utils.custom_types import (
    MinScoreOutput,
    MinScoreQASample,
    MinScoreSample,
    Sample,
)
from langtest.utils.custom_types.output import NEROutput, SequenceClassificationOutput
from langtest.utils.gender_classifier import GenderClassifier
from .utils import RepresentationOperation
from .constants import (
    default_ehtnicity_representation,
    default_economic_country_representation,
    default_religion_representation,
)


class BaseRepresentation(ABC):
    """Abstract base class for implementing representation measures.

    Attributes:
        alias_name (str): A name or list of names that identify the representation measure.
        supported_tasks (List[str]): name of the supported task for the representation measure
    Methods:
        transform(data: List[Sample]) -> Any: Transforms the input data into an output
        based on the implemented representation measure.
    """

    alias_name = None
    supported_tasks = [
        "ner",
        "text-classification",
        "question-answering",
        "summarization",
        "toxicity",
        "translation",
    ]

    @classmethod
    @abstractmethod
    def transform(
        cls, test: str, data: List[Sample], params: Dict
    ) -> Union[List[MinScoreQASample], List[MinScoreSample]]:
        """Abstract method that implements the representation measure.

        Args:
            test (str): name of the test to perform
            data (List[Sample]): The input data to be transformed.
            params (Dict): parameters for tests configuration
        Returns:
            Any: The transformed data based on the implemented representation measure.
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    async def run(
        cls, sample_list: List[Sample], model: ModelFactory, **kwargs
    ) -> List[Sample]:
        """Computes the score for the given data.

        Args:
            sample_list (List[Sample]): The input data to be transformed.
            model (ModelFactory): The model to be used for the computation.

        Returns:
            List[Sample]: The transformed samples.
        """
        raise NotImplementedError()

    @classmethod
    async def async_run(cls, sample_list: List[Sample], model: ModelFactory, **kwargs):
        """Creates a task for the run method.

        Args:
            sample_list (List[Sample]): The input data to be evaluated for representation test.
            model (ModelFactory): The model to be used for the computation.

        Returns:
            asyncio.Task: The task for the run method.
        """
        created_task = asyncio.create_task(cls.run(sample_list, model, **kwargs))
        return created_task


class GenderRepresentation(BaseRepresentation):
    """Subclass of BaseRepresentation that implements the gender representation test.

    Attributes:
        alias_name (List[str]): The list of test names that identify the representation measure.
        supported_tasks (List[str]): name of the supported task for the representation measure
    """

    alias_name = [
        "min_gender_representation_count",
        "min_gender_representation_proportion",
    ]

    @classmethod
    def transform(
        cls, test: str, data: List[Sample], params: Dict
    ) -> Union[List[MinScoreQASample], List[MinScoreSample]]:
        """Compute the gender representation measure

        Args:
            test (str): name of the test
            data (List[Sample]): The input data to be evaluated for representation test.
            params : parameters specified in config.

        Raises:
            ValueError: If sum of specified proportions in config is greater than 1

        Returns:
            Union[List[MinScoreQASample], List[MinScoreSample]]: Gender Representation test results.
        """
        assert (
            test in cls.alias_name
        ), f"Parameter 'test' should be in: {cls.alias_name}, got '{test}'"

        samples = []
        if test == "min_gender_representation_count":
            if isinstance(params["min_count"], dict):
                min_counts = params["min_count"]
            else:
                min_counts = {
                    "male": params["min_count"],
                    "female": params["min_count"],
                    "unknown": params["min_count"],
                }

            for key, value in min_counts.items():
                if hasattr(data[0], "task") and data[0].task == "question-answering":
                    sample = MinScoreQASample(
                        original_question=key,
                        original_context="-",
                        perturbed_question="-",
                        perturbed_context="-",
                        category="representation",
                        test_type="min_gender_representation_count",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    samples.append(sample)

                else:
                    sample = MinScoreSample(
                        original="-",
                        category="representation",
                        test_type="min_gender_representation_count",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    samples.append(sample)
        else:
            min_proportions = {"male": 0.26, "female": 0.26, "unknown": 0.26}

            if isinstance(params["min_proportion"], dict):
                min_proportions = params["min_proportion"]
                if sum(min_proportions.values()) > 1:
                    raise ValueError(
                        "Sum of proportions cannot be greater than 1. "
                        "So min_gender_representation_proportion test cannot run."
                    )

            for key, value in min_proportions.items():
                if hasattr(data[0], "task") and data[0].task == "question-answering":
                    sample = MinScoreQASample(
                        original_question=key,
                        original_context="-",
                        perturbed_question="-",
                        perturbed_context="-",
                        category="representation",
                        test_type="min_gender_representation_proportion",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    samples.append(sample)

                else:
                    sample = MinScoreSample(
                        original="-",
                        category="representation",
                        test_type="min_gender_representation_proportion",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    samples.append(sample)
        return samples

    @staticmethod
    async def run(
        sample_list: List[Sample], model: ModelFactory, **kwargs
    ) -> List[Sample]:
        """Computes the actual results for the Gender Representation test.

        Args:
            sample_list (List[Sample]): The input data to be evaluated for representation test.
            model (ModelFactory): The model factory object.

        Returns:
            List[Sample]: The list of samples with actual results.

        """

        progress = kwargs.get("progress_bar", False)
        classifier = GenderClassifier()
        for sample in kwargs["raw_data"]:
            if sample.task == "question-answering":
                if "perturbed_context" in sample.__annotations__:
                    genders = [
                        classifier.predict(sample.original_context)
                        for sample in kwargs["raw_data"]
                    ]
                else:
                    genders = [
                        classifier.predict(sample.original_question)
                        for sample in kwargs["raw_data"]
                    ]

            else:
                genders = [
                    classifier.predict(sample.original) for sample in kwargs["raw_data"]
                ]

        gender_counts = {
            "male": len([x for x in genders if x == "male"]),
            "female": len([x for x in genders if x == "female"]),
            "unknown": len([x for x in genders if x == "unknown"]),
        }

        total_samples = len(kwargs["raw_data"])

        for sample in sample_list:
            if progress:
                progress.update(1)

            if sample.test_type == "min_gender_representation_proportion":
                sample.actual_results = MinScoreOutput(
                    min_score=round(gender_counts[sample.test_case] / total_samples, 2)
                )
                sample.state = "done"

            elif sample.test_type == "min_gender_representation_count":
                sample.actual_results = MinScoreOutput(
                    min_score=gender_counts[sample.test_case]
                )
                sample.state = "done"
        return sample_list


class EthnicityRepresentation(BaseRepresentation):
    """Subclass of BaseRepresentation that implements the ethnicity representation test.

    Attributes:
        alias_name (List[str]): The list of test names that identify the representation measure.
        supported_tasks (List[str]): name of the supported task for the representation measure
    """

    alias_name = [
        "min_ethnicity_name_representation_count",
        "min_ethnicity_name_representation_proportion",
    ]

    @classmethod
    def transform(
        cls, test: str, data: List[Sample], params: Dict
    ) -> Union[List[MinScoreQASample], List[MinScoreSample]]:
        """Compute the ethnicity representation measure

        Args:
            test (str): name of the test
            data (List[Sample]): The input data to be evaluated for representation test.
            params : parameters specified in config.

        Raises:
            ValueError: If sum of specified proportions in config is greater than 1

        Returns:
            Union[List[MinScoreQASample], List[MinScoreSample]]: Ethnicity Representation test results.
        """
        assert (
            test in cls.alias_name
        ), f"Parameter 'test' should be in: {cls.alias_name}, got '{test}'"

        sample_list = []
        if test == "min_ethnicity_name_representation_count":
            if not params:
                expected_representation = {
                    "black": 10,
                    "asian": 10,
                    "white": 10,
                    "native_american": 10,
                    "hispanic": 10,
                    "inter_racial": 10,
                }

            else:
                if isinstance(params["min_count"], dict):
                    expected_representation = params["min_count"]

                elif isinstance(params["min_count"], int):
                    expected_representation = {
                        key: params["min_count"]
                        for key in default_ehtnicity_representation
                    }

            for key, value in expected_representation.items():
                if hasattr(data[0], "task") and data[0].task == "question-answering":
                    sample = MinScoreQASample(
                        original_question=key,
                        original_context="-",
                        perturbed_question="-",
                        perturbed_context="-",
                        category="representation",
                        test_type="min_ethnicity_name_representation_count",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)

                else:
                    sample = MinScoreSample(
                        original="-",
                        category="representation",
                        test_type="min_ethnicity_name_representation_count",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)

        else:
            if not params:
                expected_representation = {
                    "black": 0.13,
                    "asian": 0.13,
                    "white": 0.13,
                    "native_american": 0.13,
                    "hispanic": 0.13,
                    "inter_racial": 0.13,
                }

            else:
                if isinstance(params["min_proportion"], dict):
                    expected_representation = params["min_proportion"]

                    if sum(expected_representation.values()) > 1:
                        print(
                            """Sum of proportions cannot be greater than 1.
                            So min_ethnicity_name_representation_proportion test cannot run"""
                        )
                        raise ValueError()

                elif isinstance(params["min_proportion"], float):
                    expected_representation = {
                        key: params["min_proportion"]
                        for key in default_ehtnicity_representation
                    }
                    if sum(expected_representation.values()) > 1:
                        print(
                            """Sum of proportions cannot be greater than 1.
                        So min_ethnicity_name_representation_proportion test cannot run"""
                        )
                        raise ValueError()

            for key, value in expected_representation.items():
                if hasattr(data[0], "task") and data[0].task == "question-answering":
                    sample = MinScoreQASample(
                        original_question=key,
                        original_context="-",
                        perturbed_question="-",
                        perturbed_context="-",
                        category="representation",
                        test_type="min_ethnicity_name_representation_proportion",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)

                else:
                    sample = MinScoreSample(
                        original="-",
                        category="representation",
                        test_type="min_ethnicity_name_representation_proportion",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)

        return sample_list

    @staticmethod
    async def run(
        sample_list: List[Sample], model: ModelFactory, **kwargs
    ) -> List[Sample]:
        """Computes the actual results for the ethnicity representation test.

        Args:
            sample_list (List[Sample]): The input data to be evaluated for representation test.
            model (ModelFactory): The model to be used for evaluation.

        Returns:
            List[Sample]: The list of samples with actual results.
        """
        progress = kwargs.get("progress_bar", False)

        entity_representation = RepresentationOperation.get_ethnicity_representation_dict(
            kwargs["raw_data"]
        )

        for sample in sample_list:
            if sample.test_type == "min_ethnicity_name_representation_proportion":
                entity_representation_proportion = (
                    RepresentationOperation.get_entity_representation_proportions(
                        entity_representation
                    )
                )
                actual_representation = {
                    **default_ehtnicity_representation,
                    **entity_representation_proportion,
                }

                sample.actual_results = MinScoreOutput(
                    min_score=round(actual_representation[sample.test_case], 2)
                )
                sample.state = "done"

            elif sample.test_type == "min_ethnicity_name_representation_count":
                actual_representation = {
                    **default_ehtnicity_representation,
                    **entity_representation,
                }
                sample.actual_results = MinScoreOutput(
                    min_score=round(actual_representation[sample.test_case], 2)
                )
                sample.state = "done"

            if progress:
                progress.update(1)

        return sample_list


class LabelRepresentation(BaseRepresentation):
    """Subclass of BaseRepresentation that implements the label representation test.

    Attributes:
        alias_name (List[str]): The list of test names that identify the representation measure.
        supported_tasks (List[str]): name of the supported task for the representation measure
    """

    alias_name = ["min_label_representation_count", "min_label_representation_proportion"]

    supported_tasks = ["ner", "text-classification"]

    @classmethod
    def transform(
        cls, test: str, data: List[Sample], params: Dict
    ) -> Union[List[MinScoreQASample], List[MinScoreSample]]:
        """Compute the label representation measure

        Args:
            test (str): name of the test
            data (List[Sample]): The input data to be evaluated for representation test.
            params (Dict): parameters specified in config.

        Raises:
            ValueError: If sum of specified proportions in config is greater than 1

        Returns:
            Union[List[MinScoreQASample], List[MinScoreSample]]: Label Representation test results.
        """
        assert (
            test in cls.alias_name
        ), f"Parameter 'test' should be in: {cls.alias_name}, got '{test}'"

        sample_list = []
        labels = [s.expected_results.predictions for s in data]
        if isinstance(data[0].expected_results, NEROutput):
            labels = [x.entity.split("-")[-1] for sentence in labels for x in sentence]
        elif isinstance(data[0].expected_results, SequenceClassificationOutput):
            labels = [x.label for sentence in labels for x in sentence]
        labels = set(labels)

        if test == "min_label_representation_count":
            if not params:
                expected_representation = {k: 10 for k in labels}

            else:
                if isinstance(params["min_count"], dict):
                    expected_representation = params["min_count"]

                elif isinstance(params["min_count"], int):
                    expected_representation = {key: params["min_count"] for key in labels}

            for key, value in expected_representation.items():
                sample = MinScoreSample(
                    original="-",
                    category="representation",
                    test_type="min_label_representation_count",
                    test_case=key,
                    expected_results=MinScoreOutput(min_score=value),
                )
                sample_list.append(sample)

        else:
            if not params:
                expected_representation = {k: (1 / len(k)) * 0.8 for k in labels}

            else:
                if isinstance(params["min_proportion"], dict):
                    expected_representation = params["min_proportion"]

                    if sum(expected_representation.values()) > 1:
                        print(
                            """Sum of proportions cannot be greater than 1.
                        So min_label_representation_proportion test cannot run"""
                        )
                        raise ValueError()

                elif isinstance(params["min_proportion"], float):
                    expected_representation = {
                        key: params["min_proportion"] for key in labels
                    }
                    if sum(expected_representation.values()) > 1:
                        print(
                            """Sum of proportions cannot be greater than 1.
                        So min_label_representation_proportion test cannot run"""
                        )
                        raise ValueError()

            for key, value in expected_representation.items():
                sample = MinScoreSample(
                    original="-",
                    category="representation",
                    test_type=test,
                    test_case=key,
                    expected_results=MinScoreOutput(min_score=value),
                )
                sample_list.append(sample)

        return sample_list

    @staticmethod
    async def run(
        sample_list: List[Sample], model: ModelFactory, **kwargs
    ) -> List[Sample]:
        """Computes the actual representation of the labels in the dataset.

        Args:
            sample_list (List[Sample]): The input data to be evaluated for representation test.
            model (ModelFactory): The model to be evaluated.

        Returns:
            List[Sample]: Label Representation test results.
        """
        progress = kwargs.get("progress_bar", False)

        entity_representation = RepresentationOperation.get_label_representation_dict(
            kwargs["raw_data"]
        )

        for sample in sample_list:
            if progress:
                progress.update(1)

            if sample.test_type == "min_label_representation_proportion":
                entity_representation_proportion = (
                    RepresentationOperation.get_entity_representation_proportions(
                        entity_representation
                    )
                )
                actual_representation = {**entity_representation_proportion}
                sample.actual_results = MinScoreOutput(
                    min_score=round(actual_representation[sample.test_case], 2)
                )
                sample.state = "done"
            elif sample.test_type == "min_label_representation_count":
                actual_representation = {**entity_representation}
                sample.actual_results = MinScoreOutput(
                    min_score=round(actual_representation[sample.test_case], 2)
                )
                sample.state = "done"

        return sample_list


class ReligionRepresentation(BaseRepresentation):
    """Subclass of BaseRepresentation that implements the religion representation test.

    Attributes:
        alias_name (List[str]): The list of test names that identify the representation measure.
        supported_tasks (List[str]): name of the supported task for the representation measure
    """

    alias_name = [
        "min_religion_name_representation_count",
        "min_religion_name_representation_proportion",
    ]
    supported_tasks = [
        "ner",
        "text-classification",
        "question-answering",
        "summarization",
    ]

    @classmethod
    def transform(
        cls, test: str, data: List[Sample], params: Dict
    ) -> Union[List[MinScoreQASample], List[MinScoreSample]]:
        """Compute the religion representation measure

        Args:
            test (str): name of the test
            data (List[Sample]): The input data to be evaluated for representation test.
            params : parameters specified in config.

        Raises:
            ValueError: If sum of specified proportions in config is greater than 1

        Returns:
            Union[List[MinScoreQASample], List[MinScoreSample]]: Religion Representation test results.
        """
        assert (
            test in cls.alias_name
        ), f"Parameter 'test' should be in: {cls.alias_name}, got '{test}'"

        sample_list = []
        if test == "min_religion_name_representation_count":
            if not params:
                expected_representation = {
                    "muslim": 5,
                    "hindu": 5,
                    "sikh": 5,
                    "christian": 5,
                    "jain": 5,
                    "buddhist": 5,
                    "parsi": 5,
                }

            else:
                if isinstance(params["min_count"], dict):
                    expected_representation = params["min_count"]

                elif isinstance(params["min_count"], int):
                    expected_representation = {
                        key: params["min_count"]
                        for key in default_religion_representation
                    }

            for key, value in expected_representation.items():
                if hasattr(data[0], "task") and data[0].task == "question-answering":
                    sample = MinScoreQASample(
                        original_question=key,
                        original_context="-",
                        perturbed_question="-",
                        perturbed_context="-",
                        category="representation",
                        test_type="min_religion_name_representation_count",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)
                else:
                    sample = MinScoreSample(
                        original="-",
                        category="representation",
                        test_type="min_religion_name_representation_count",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)

        else:
            if not params:
                expected_representation = {
                    "muslim": 0.11,
                    "hindu": 0.11,
                    "sikh": 0.11,
                    "christian": 0.11,
                    "jain": 0.11,
                    "buddhist": 0.11,
                    "parsi": 0.11,
                }

            else:
                if isinstance(params["min_proportion"], dict):
                    expected_representation = params["min_proportion"]

                    if sum(expected_representation.values()) > 1:
                        print(
                            """Sum of proportions cannot be greater than 1.
                        So min_religion_name_representation_proportion test cannot run"""
                        )
                        raise ValueError()

                elif isinstance(params["min_proportion"], float):
                    expected_representation = {
                        key: params["min_proportion"]
                        for key in default_religion_representation
                    }
                    if sum(expected_representation.values()) > 1:
                        print(
                            """Sum of proportions cannot be greater than 1.
                        So min_religion_name_representation_proportion test cannot run"""
                        )
                        raise ValueError()

            entity_representation = (
                RepresentationOperation.get_religion_name_representation_dict(data)
            )
            entity_representation_proportion = (
                RepresentationOperation.get_entity_representation_proportions(
                    entity_representation
                )
            )
            actual_representation = {
                **default_religion_representation,
                **entity_representation_proportion,
            }
            for key, value in expected_representation.items():
                if hasattr(data[0], "task") and data[0].task == "question-answering":
                    sample = MinScoreQASample(
                        original_question=key,
                        original_context="-",
                        perturbed_question="-",
                        perturbed_context="-",
                        category="representation",
                        test_type="min_religion_name_representation_proportion",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)
                else:
                    sample = MinScoreSample(
                        original="-",
                        category="representation",
                        test_type="min_religion_name_representation_proportion",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                        actual_results=MinScoreOutput(
                            min_score=actual_representation[key]
                        ),
                        state="done",
                    )
                    sample_list.append(sample)

        return sample_list

    @staticmethod
    async def run(
        sample_list: List[Sample], model: ModelFactory, **kwargs
    ) -> List[Sample]:
        """Computes the actual representation of religion names in the data.

        Args:
            sample_list (List[Sample]): The input data to be evaluated for representation test.
            model (ModelFactory): The model to be evaluated.

        Returns:
            List[Sample]: Religion Representation test results.

        """
        progress = kwargs.get("progress_bar", False)

        entity_representation = (
            RepresentationOperation.get_religion_name_representation_dict(
                kwargs["raw_data"]
            )
        )

        for sample in sample_list:
            if sample.test_type == "min_religion_name_representation_proportion":
                entity_representation_proportion = (
                    RepresentationOperation.get_entity_representation_proportions(
                        entity_representation
                    )
                )
                actual_representation = {
                    **default_religion_representation,
                    **entity_representation_proportion,
                }

                sample.actual_results = MinScoreOutput(
                    min_score=round(actual_representation[sample.test_case], 2)
                )
                sample.state = "done"

            elif sample.test_type == "min_religion_name_representation_count":
                actual_representation = {
                    **default_religion_representation,
                    **entity_representation,
                }

                sample.actual_results = MinScoreOutput(
                    min_score=round(actual_representation[sample.test_case], 2)
                )
                sample.state = "done"

            if progress:
                progress.update(1)

        return sample_list


class CountryEconomicRepresentation(BaseRepresentation):
    """Subclass of BaseRepresentation that implements the country economic representation test.

    Attributes:
        alias_name (List[str]): The list of test names that identify the representation measure.
        supported_tasks (List[str]): name of the supported task for the representation measure
    """

    alias_name = [
        "min_country_economic_representation_count",
        "min_country_economic_representation_proportion",
    ]

    @classmethod
    def transform(
        cls, test: str, data: List[Sample], params: Dict
    ) -> Union[List[MinScoreQASample], List[MinScoreSample]]:
        """Compute the country economic representation measure

        Args:
            test (str): name of the test
            data (List[Sample]): The input data to be evaluated for representation test.
            params : parameters specified in config.

        Raises:
            ValueError: If sum of specified proportions in config is greater than 1

        Returns:
            Union[List[MinScoreQASample], List[MinScoreSample]]: Country Economic Representation test results.
        """
        sample_list = []

        if test == "min_country_economic_representation_count":
            if not params:
                expected_representation = {
                    "high_income": 10,
                    "low_income": 10,
                    "lower_middle_income": 10,
                    "upper_middle_income": 10,
                }

            else:
                if isinstance(params["min_count"], dict):
                    expected_representation = params["min_count"]

                elif isinstance(params["min_count"], int):
                    expected_representation = {
                        key: params["min_count"]
                        for key in default_economic_country_representation
                    }

            for key, value in expected_representation.items():
                if hasattr(data[0], "task") and data[0].task == "question-answering":
                    sample = MinScoreQASample(
                        original_question=key,
                        original_context="-",
                        perturbed_question="-",
                        perturbed_context="-",
                        category="representation",
                        test_type="min_country_economic_representation_count",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)
                else:
                    sample = MinScoreSample(
                        original="-",
                        category="representation",
                        test_type="min_country_economic_representation_count",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)

        else:
            if not params:
                expected_representation = {
                    "high_income": 0.20,
                    "low_income": 0.20,
                    "lower_middle_income": 0.20,
                    "upper_middle_income": 0.20,
                }

            else:
                if isinstance(params["min_proportion"], dict):
                    expected_representation = params["min_proportion"]

                    if sum(expected_representation.values()) > 1:
                        print(
                            """Sum of proportions cannot be greater than 1.
                        So min_country_economic_representation_proportion test cannot run"""
                        )
                        raise ValueError()

                elif isinstance(params["min_proportion"], float):
                    expected_representation = {
                        key: params["min_proportion"]
                        for key in default_economic_country_representation
                    }
                    if sum(expected_representation.values()) > 1:
                        print(
                            "Sum of proportions cannot be greater than 1. So "
                            "min_country_economic_representation_proportion test cannot run"
                        )
                        raise ValueError()

            for key, value in expected_representation.items():
                if hasattr(data[0], "task") and data[0].task == "question-answering":
                    sample = MinScoreQASample(
                        original_question=key,
                        original_context="-",
                        perturbed_question="-",
                        perturbed_context="-",
                        category="representation",
                        test_type="min_country_economic_representation_proportion",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)

                else:
                    sample = MinScoreSample(
                        original="-",
                        category="representation",
                        test_type="min_country_economic_representation_proportion",
                        test_case=key,
                        expected_results=MinScoreOutput(min_score=value),
                    )
                    sample_list.append(sample)

        return sample_list

    @staticmethod
    async def run(
        sample_list: List[Sample], model: ModelFactory, **kwargs
    ) -> List[Sample]:
        """Computes the actual results for the country economic representation test.

        Args:
            sample_list (List[Sample]): The input data to be evaluated for representation test.
            model (ModelFactory): The model to be used for evaluation.

        Returns:
            List[Sample]: Country Economic Representation test results.

        """
        progress = kwargs.get("progress_bar", False)

        entity_representation = (
            RepresentationOperation.get_country_economic_representation_dict(
                kwargs["raw_data"]
            )
        )

        for sample in sample_list:
            if sample.test_type == "min_country_economic_representation_proportion":
                entity_representation_proportion = (
                    RepresentationOperation.get_entity_representation_proportions(
                        entity_representation
                    )
                )
                actual_representation = {
                    **default_economic_country_representation,
                    **entity_representation_proportion,
                }

                sample.actual_results = MinScoreOutput(
                    min_score=round(actual_representation[sample.test_case], 2)
                )
                sample.state = "done"

            elif sample.test_type == "min_country_economic_representation_count":
                actual_representation = {
                    **default_economic_country_representation,
                    **entity_representation,
                }

                sample.actual_results = MinScoreOutput(
                    min_score=round(actual_representation[sample.test_case], 2)
                )
                sample.state = "done"

            if progress:
                progress.update(1)

        return sample_list
